#  Copyright (c) 2022 Christoph Finkensiep

import abc
import numbers
import numpy as np
import copy

from pitchtypes.basetypes import Pitch, Interval, Diatonic, Chromatic
from pitchtypes.spelled import Spelled, SpelledInterval, SpelledIntervalClass, SpelledPitch, SpelledPitchClass

class SpelledArray(abc.ABC):
    """
    A common base class for vectorized spelled pitch and interval types.
    Uses the same interface as the Spelled types.
    """

    # constructors

    @abc.abstractstaticmethod
    def from_onehot():
        raise NotImplementedError

    # printing

    _print_name = "SpelledArray"
    
    def __repr__(self):
        # For vectorized types, name() returns an array of names,
        # so we convert it to a string here:
        return f"{self._print_name}({np.array2string(self.name(), separator=', ')})"

    def __str__(self):
        # with np.printoptions(formatter={'all': lambda x: str(x)}):
        #     return f"{self._print_name}{self.name()}"
        return np.array2string(self.name(), formatter={'all': lambda x: str(x)})

    # mandatory array methods

    @abc.abstractmethod
    def __copy__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __deepcopy__(self, memo):
        raise NotImplementedError

    def copy(self):
        """
        Returns a shallow copy of the array.
        This also creates copies of the underlying numpy arrays.

        :return: a copy of the array
        """
        return copy.copy(self)

    def deepcopy(self):
        """
        Returns a deep copy of the array.

        :return: a deepcopy of the array
        """
        return copy.deepcopy(self)

    @abc.abstractmethod
    def __getitem__(self, index):
        """
        Returns an item or a subarray of the spelled array.
        Supports advanced indexing as on numpy arrays.

        :param index: a numpy-compatible index into the array
        :return: a new array or a scalar value (depending on the index)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __setitem__(self, index, item):
        """
        Sets the given indices to the given item(s).
        Supports advanced indexing as on numpy arrays.

        :param index: a numpy-compatible index into the array
        :param item: an array or a scalar to assign to the indicated sub-array
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __contains__(self, item):
        """
        Returns true if the array contains the given interval/pitch.

        :param item: the potential item to test
        :return: ``True`` if the array contains ``item``, otherwise ``False``
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        """
        Returns the length of the array (first dimension, as in numpy).

        :return: the length of the array (1st dimension, integer)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError

    def array_equal(self, other):
        """
        Returns True if self and other are the equal,
        False otherwise.

        :param other: another spelled array of the same type
        :return: ``True`` if the two arrays are equal, ``False`` otherwise
        """
        try:
            return (self.compare(other) == 0).all()
        except TypeError:
            return False

    # element-wise comparison
    
    @abc.abstractmethod
    def compare(self, other):
        """
        Element-wise comparison between two spelled arrays.

        Returns 0 where the elements are equal,
        1 where the first element is greater,
        and -1 where the second element is greater.

        The respective ordering differs between types.
        Non-class pitches and intervals use diatonic ordering,
        interval/pitch classes use line-of-fifths ordering.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).
        To test the overall equality of two spelled arrays,
        use :py:meth:`array_equal <SpelledArray.array_equal>`

        :param other: another spelled array or scalar
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        raise NotImplementedError
    
    def __lt__(self, other):
        """
        Element-wise ``<`` between two spelled arrays.

        :param other: the spelled array or scalar to compare to
        :return: a boolean array that indicates where this array is ``< other``
        """
        try:
            return self.compare(other) == -1
        except TypeError:
           return NotImplemented

    def __le__(self, other):
        """
        Element-wise ``<=`` between two spelled arrays.

        :param other: the spelled array or scalar to compare to
        :return: a boolean array that indicates where this array is ``<= other``
        """
        try:
            return self.compare(other) != 1
        except TypeError:
           return NotImplemented

    def __gt__(self, other):
        """
        Element-wise ``>`` between two spelled arrays.

        :param other: the spelled array or scalar to compare to
        :return: a boolean array that indicates where this array is ``> other``
        """
        try:
            return self.compare(other) == 1
        except TypeError:
           return NotImplemented

    def __ge__(self, other):
        """
        Element-wise ``>=`` between two spelled arrays.

        :param other: the spelled array or scalar to compare to
        :return: a boolean array that indicates where this array is ``>= other``
        """
        try:
            return self.compare(other) != -1
        except TypeError:
           return NotImplemented

    def __eq__(self, other):
        """
        Element-wise ``<=`` between two spelled arrays.

        :param other: the spelled array or scalar to compare to
        :return: a boolean array that indicates where this array is ``== other``
        """
        try:
            return self.compare(other) == 0
        except TypeError:
           return NotImplemented

    def __ne__(self, other):
        """
        Element-wise ``<=`` between two spelled arrays.

        :param other: the spelled array or scalar to compare to
        :return: a boolean array that indicates where this array is ``!= other``
        """
        try:
            return self.compare(other) != 0
        except TypeError:
           return NotImplemented

    # spelled interface

    @abc.abstractmethod
    def name(self):
        """
        Returns the names of the objects in the array
        as an array of strings of the same shape.

        :return: an array of notation strings
        """
        raise NotImplementedError

    @abc.abstractmethod
    def fifths(self):
        """
        Return the position of the interval on the line of fifths.

        :return: an array of fifths (integers)
        """
        raise NotImplementedError
    
    @abc.abstractmethod
    def octaves(self):
        """
        For intervals, return the number of octaves the interval spans.
        Negative intervals start with -1, decreasing.
        For pitches, return the absolute octave of the pitch.

        :return: an array of external/independent octaves (integers)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def internal_octaves(self):
        """
        Return the internal octave representation of a pitch,
        which is dependent on the fifths.

        Only use this if you know what you are doing.

        :return: an array of internal/dependent octaves (integers)
        """
        raise NotImplementedError

    def degree(self):
        """
        Return the "relative scale degree" (0-6) to which the interval points
        (unison=0, 2nd=1, octave=0, 2nd down=6, etc.).
        For pitches, return the integer that corresponds to the letter (C=0, D=1, ...).

        :return: an array of degrees (integers)
        """
        return (self.fifths() * 4) % 7

    @abc.abstractmethod
    def alteration(self):
        """
        Return the number of semitones by which the interval is altered
        from its the perfect or major variant.
        Positive alteration always indicates augmentation,
        negative alteration indicates diminution (minor or smaller) of the interval's magnitude.
        For interval classes, alteration refers to the upward version of the interval
        (e.g. for ``m7``/``-M2`` it is -1).
        For pitches, return the accidentals (positive=sharps, negative=flats, 0=natural).

        :return: an array of alterations (integers)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def onehot(self):
        """
        Return a one-hot encoded tensor representing the elements of the array.
        Specialized versions of this method take ranges for their respective dimensions.

        :return: a one-hot tensor for the array
        """
        raise NotImplementedError

class AbstractSpelledArrayInterval(abc.ABC):
    """
    The interface for spelled interval array types.
    """
    @abc.abstractmethod
    def generic(self):
        """
        Return the generic interval, i.e. the number of diatonic steps modulo octave.
        Unlike degree(), the result respects the sign of the interval
        (unison=0, 2nd up=1, 2nd down=-1).

        :return: an array of generic intervals (integers)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def diatonic_steps(self):
        """
        Return the diatonic steps of the interval (unison=0, 2nd=1, ..., octave=7, ...).
        Respects both direction and octaves.

        :return: an array of diatonic steps (integers)
        """
        raise NotImplementedError

class AbstractSpelledArrayPitch(abc.ABC):
    """
    The interface for spelled pitch array types.
    """
    @abc.abstractmethod
    def letter(self):        
        """
        Returns the letter associated with the pitch (without accidentals).

        :return: an array of pitch letters (single-character strings)
        """
        raise NotImplementedError
        

class SpelledIntervalArray(SpelledArray, AbstractSpelledArrayInterval, Interval, Diatonic, Chromatic):
    """
    Represents an array of spelled intervals.
    """

    _print_name = "asi"
    
    def __init__(self, fifths, octaves):
        """        
        Takes two numpy arrays,
        one for fifths and one for internal/dependent octaves,
        both as integers.

        :param fifths: the internal fifths of each interval (numpy array of integers)
        :param octaves: the internal octaves of each interval (numpy array of integers)
        """
        if fifths.shape != octaves.shape:
            raise ValueError(f"Cannot create SpelledIntervalArray from arrays of different sizes ({fifths.shape} and {octaves.shape}).")
        self._fifths = fifths
        self._octaves = octaves

    @staticmethod
    def from_independent(fifths, octaves):
        """
        Create an interval array from fifths (interval class) and independent/external octaves.

        :param fifths:  the internal fifths of each interval (numpy array of integers)
        :param octaves: the external/independent octaves of each interval (numpy array of integers)
        :return: the corresponding interval array
        """
        return SpelledIntervalArray(fifths, octaves - (fifths * 4) // 7)

    @staticmethod
    def from_strings(strings):
        """
        Create an interval array from an array of strings (using spelled interval notation).

        :param strings: an array-like of interval notation strings
        :return: the corresponding interval array
        """
        def parse_interval(string):
            sign, octave, fifth = Spelled.parse_interval(string)
            if octave is None:
                raise ValueError(f"Missing octave specifier in interval '{string}'.")
            return sign, octave, fifth
        sign, octaves, fifths = np.vectorize(parse_interval, otypes=[np.int_, np.int_, np.int_])(strings)
        return SpelledIntervalArray(fifths * sign, (octaves - (fifths * 4) // 7) * sign)

    @staticmethod
    def from_array(intervals):
        """
        Create an interval array from an array of intervals.

        :param intervals: an array-like of ``SpelledInterval``
        :return: the corresponding interval array
        """
        def from_interval(interval):
            return interval.fifths(), interval.internal_octaves()
        fifths, octaves = np.vectorize(from_interval, otypes=[np.int_, np.int_])(intervals)
        return SpelledIntervalArray(fifths, octaves)

    @staticmethod
    def from_onehot(onehot, fifth_low, octave_low):
        """
        Create a spelled interval array from a one-hot tensor.
        ``fifth_low`` denotes the lower bound of the fifth range used in the vector,
        ``octave_low`` the lower bound of the octave range.
        The shape of the resulting array will be equivalent to the first n-2 dimensions of the input tensor.

        :param onehot: a one-hot tensor representing the intervals (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot tensor (integer)
        :param octave_low: the lowest octave expressible in the one-hot tensor (integer)
        :return: the corresponding interval array
        """
        if (onehot.sum((-2,-1)) != 1).any():
            raise ValueError(f"{onehot} is not a one-hot tensor.")
        
        ones = np.where(onehot==1)
        indices = ones[:-2] # first n-2 dimensions are indices
        fifths_values = ones[-2] + fifth_low
        octs_values = ones[-1] + octave_low

        new_shape = onehot.shape[:-2]
        new_fifths = np.zeros(new_shape, dtype=int)
        new_fifths[indices] = fifths_values
        new_octaves = np.zeros(new_shape, dtype=int)
        new_octaves[indices] = octs_values
        return SpelledIntervalArray.from_independent(new_fifths, new_octaves)

    # collection interface

    def __copy__(self):
        return SpelledIntervalArray(copy.copy(self.fifths()), copy.copy(self.internal_octaves()))

    def __deepcopy__(self, memo):
        return SpelledIntervalArray(copy.deepcopy(self.fifths(), memo), copy.deepcopy(self.internal_octaves(), memo))
    
    def __getitem__(self, index):
        f = self._fifths[index]
        o = self._octaves[index]
        if isinstance(f, numbers.Integral):
            return SpelledInterval.from_fifths_and_octaves(f, o)
        else:
            return SpelledIntervalArray(f, o)

    def __setitem__(self, index, item):
        if isinstance(item, SpelledInterval) or isinstance(item, SpelledIntervalArray):
            self._fifths[index] = item.fifths()
            self._octaves[index] = item.internal_octaves()
        else:
            raise TypeError(f"Cannot set elements of SpelledIntervalArray to {type(item)}.")

    def __contains__(self, item):
        if isinstance(item, SpelledInterval):
            return ((self.fifths() == item.fifths()) & \
                    (self.internal_octaves() == item.internal_octaves())).any()
        else:
            return False

    def __len__(self):
        return len(self.fifths())

    class SpelledIntervalArrayIter:
        def __init__(self, array):
            self._fifths_iter = array.fifths().__iter__()
            self._octaves_iter = array.internal_octaves().__iter__()

        def __next__(self):
            f = self._fifths_iter.__next__()
            o = self._octaves_iter.__next__()
            if isinstance(f, numbers.Integral):
                return SpelledInterval.from_fifths_and_octaves(f, o)
            else:
                return SpelledIntervalArray(f, o)
        
    def __iter__(self):
        return self.SpelledIntervalArrayIter(self)
        
    # interval interface

    @classmethod
    def unison(cls, shape):
        """
        Create an array of the given shape filled with perfect unisons (P1:0).

        :param shape: the shape of the resulting array (tuple of integers)
        :return: a ``SpelledIntervalArray`` of shape ``shape`` filled with P1:0
        """
        return cls(np.full(shape, 0, dtype=np.int_),np.full(shape, 0, dtype=np.int_))

    @classmethod
    def octave(cls, shape):
        """
        Return an array of the given shape filled with perfect octaves (P1:1).

        :param shape: the shape of the resulting array (tuple of integers)
        :return: a ``SpelledIntervalArray`` of shape ``shape`` filled with P1:0
        """
        return cls(np.full(shape, 0, dtype=np.int_),np.full(shape, 1, dtype=np.int_))

    @classmethod
    def chromatic_semitone(cls, shape):
        """
        Returns an array of the given shape filled with chromatic semitones (a1:0).

        :param shape: the shape of the resulting array (tuple of integers)
        :return: a ``SpelledIntervalArray`` of shape ``shape`` filled with a1:0
        """
        return SpelledIntervalArray(np.full(shape, 7, dtype=np.int_), np.full(shape, -4, dtype=np.int_))

    def __add__(self, other):
        if type(other) == SpelledInterval or type(other) == SpelledIntervalArray:
            return SpelledIntervalArray(self.fifths() + other.fifths(),
                                        self.internal_octaves() + other.internal_octaves())
        else:
            return NotImplemented

    def __sub__(self, other):
        if type(other) == SpelledInterval or type(other) == SpelledIntervalArray:
            return SpelledIntervalArray(self.fifths() - other.fifths(),
                                        self.internal_octaves() - other.internal_octaves())
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, numbers.Integral) or\
           (hasattr(other, 'dtype') and issubclass(other.dtype.type, numbers.Integral)):
            return SpelledIntervalArray(self.fifths() * other, self.internal_octaves() * other)
        else:
            return NotImplemented
    
    def __neg__(self):
        return SpelledIntervalArray(-self.fifths(), -self.internal_octaves())

    def __abs__(self):
        downs = self.direction() < 0
        abs_fifths = self.fifths().copy()
        abs_octs = self.internal_octaves().copy()
        # manually invert the intervals that point downwards
        abs_fifths[downs] = -abs_fifths[downs]
        abs_octs[downs] = -abs_octs[downs]
        return SpelledIntervalArray(abs_fifths, abs_octs)

    def direction(self):
        """
        Returns the direction of the interval (1=up / 0=neutral / -1=down).
        The perfect unisons (``P1:0``) is considered neutral.
        
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        dia = np.sign(self.diatonic_steps())
        mask = dia == 0
        dia[mask] = np.sign((self.fifths()[mask] + 1) // 7)
        return dia

    def ic(self):
        return SpelledIntervalClassArray(self.fifths())

    def embed(self):
        return self
    
    def is_step(self):
        return abs(self.diatonic_steps()) <= 1

    # spelled interface

    def name(self):
        def interval_name(fifths, octaves, direction):
            if direction == -1:
                return "-" + Spelled.interval_class_from_fifths(-fifths) + ":" + str(octaves)
            else:
                return Spelled.interval_class_from_fifths(fifths) + ":" + str(octaves)
        return np.vectorize(interval_name, otypes=[np.str_])(self.fifths(), abs(self).octaves(), self.direction())

    def compare(self, other):
        """
        Element-wise comparison between two spelled arrays.

        Returns 0 where the elements are equal,
        1 where the first element is greater,
        and -1 where the second element is greater.

        Spelled intervals use diatonic ordering,
        for example ``P4:0 < a4:0 < aaaa4:0 < d5:0 < P5:0``.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).
        To test the overall equality of two spelled arrays,
        use :py:meth:`array_equal <SpelledIntervalArray.array_equal>`

        :param other: another ``SpelledIntervalArray`` or ``SpelledInterval``
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledInterval) or isinstance(other, SpelledIntervalArray):
            return (self - other).direction()
        else:
            raise TypeError(f"Cannot elements of {type(self)} to {type(other)}.")

    def fifths(self):
        return self._fifths

    def octaves(self):
        return self.internal_octaves() + (self.fifths() * 4) // 7

    def internal_octaves(self):
        return self._octaves

    def generic(self):
        downs = self.direction() < 0
        degrees = self.degree()
        degrees[downs] = -((-degrees[downs]) % 7)
        return degrees

    def diatonic_steps(self):
        return (self.fifths() * 4) + (self.internal_octaves() * 7)

    def alteration(self):
        return (abs(self).fifths() + 1) // 7

    def onehot(self, fifth_range, octave_range, dtype=int):
        """
        Returns a one-hot encoding of the intervals in fifths and independent octaves as the innermost dimensions.
        The range of fifths and octaves is given by ``fifth_range`` and ``octave_range`` respectively,
        where each is a tuple ``(lower, upper)``.
        The outer shape of the output tensor is identical to the shape of the original array,
        so the resulting shape is ``original_shape + (n_fifths, n_octaves)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param octave_range: the (inclusive) range of octaves (pair of integers)
        :param dtype: dtype of the resulting array (default: ``int``)
        :return: a one-hot tensor (numpy array)
        """
        flow, fhigh = fifth_range
        olow, ohigh = octave_range
        f = self.fifths()
        o = self.octaves()
        if (f < flow).any() or (f > fhigh).any():
            raise ValueError(f"The interval {self} is outside the given fifth range {fifth_range}.")
        if (o < olow).any() or (o > ohigh).any():
            raise ValueError(f"The interval {self} is outside the given octave range {octave_range}.")
        # translate fifths and octaves to 0-based indices (from [low,high] to [0,high-low])
        f = f - flow
        o = o - olow

        # compute one-hot indices
        inner_shape = f.shape
        # a tuple of indices:
        # the first dimensions encode the position in the original array,
        # the last two dimensions encode the fifth and octave position.
        indices = tuple(np.indices(inner_shape)) + (f, o)

        # initialize one-hot tensor with 0s
        out = np.zeros(inner_shape + (fhigh-flow+1, ohigh-olow+1), dtype=dtype)
        # set all elements picked by the indices to 1
        out[indices] = 1
        return out

class SpelledIntervalClassArray(SpelledArray, AbstractSpelledArrayInterval, Interval, Diatonic, Chromatic):
    """
    Represents an array of spelled interval classes, i.e. intervals without octave information.
    """

    _print_name = "asic"
    
    def __init__(self, fifths):
        """
        Takes a numpy array of fifths as integers.

        :param fifths: the internal fifths of each interval class (numpy array of integers)
        """
        self._fifths = fifths

    @staticmethod
    def from_strings(strings):
        """
        Create an interval-class array from an array of strings.

        :param strings: an array-like of interval-class notation strings
        :return: the corresponding interval-class array
        """
        def parse_ic(string):
            sign, octave, fifth = Spelled.parse_interval(string)
            if octave is not None:
                raise ValueError(f"Interval classes cannot have octave specifiers ({string}).")
            return sign, fifth
        sign, fifths = np.vectorize(parse_ic, otypes=[np.int_, np.int_])(strings)
        return SpelledIntervalClassArray(fifths * sign)

    @staticmethod
    def from_array(intervals):
        """
        Create an interval class array from an array of interval classes.

        :param intervals: an array-like of ``SpelledIntervalClass``
        :return: the corresponding interval-class array
        """
        fifths = np.vectorize(lambda i: i.fifths(), otypes=[np.int_])(intervals)
        return SpelledIntervalClassArray(fifths)

    @staticmethod
    def from_onehot(onehot, fifth_low):
        """
        Create a spelled interval-class array from a one-hot tensor.
        ``fifth_low`` denotes the lower bound of the fifth range used in the vector.
        The shape of the resulting array will be equivalent to the first n-1 dimensions of the input tensor.

        :param onehot: a one-hot tensor representing the interval classes (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot tensor (integer)
        :return: the corresponding interval-class array
        """
        if (onehot.sum(-1) != 1).any():
            raise ValueError(f"{onehot} is not a one-hot tensor.")
        
        ones = np.where(onehot==1)
        indices = ones[:-1] # first n-1 dimensions are indices
        fifths_values = ones[-1] + fifth_low

        new_shape = onehot.shape[:-1]
        new_fifths = np.zeros(new_shape, dtype=int)
        new_fifths[indices] = fifths_values
        return SpelledIntervalClassArray(new_fifths)

    # collection interface

    def __copy__(self):
        return SpelledIntervalClassArray(copy.copy(self.fifths()))

    def __deepcopy__(self, memo):
        return SpelledIntervalClassArray(copy.deepcopy(self.fifths(), memo))
    
    def __getitem__(self, index):
        f = self._fifths[index]
        if isinstance(f, numbers.Integral):
            return SpelledIntervalClass.from_fifths(f)
        else:
            return SpelledIntervalClassArray(f)

    def __setitem__(self, index, item):
        if isinstance(item, SpelledIntervalClass) or isinstance(item, SpelledIntervalClassArray):
            self._fifths[index] = item.fifths()
        else:
            raise TypeError(f"Cannot set elements of SpelledIntervalClassArray to {type(item)}.")

    def __contains__(self, item):
        if isinstance(item, SpelledIntervalClass):
            return item.fifths() in self.fifths()
        else:
            return False

    def __len__(self):
        return len(self.fifths())
        
    class SpelledIntervalClassArrayIter:
        def __init__(self, array):
            self._fifths_iter = array.fifths().__iter__()
            
        def __next__(self):
            f = self._fifths_iter.__next__()
            if isinstance(f, numbers.Integral):
                return SpelledIntervalClass.from_fifths(f)
            else:
                return SpelledIntervalClassArray(f)
        
    def __iter__(self):
        return self.SpelledIntervalClassArrayIter(self)
        
    # interval interface

    @classmethod
    def unison(cls, shape):
        """
        Return an array of the given shape filled with perfect unisons (P1).

        :param shape: the shape of the resulting array (tuple of integers)
        :return: a ``SpelledIntervalClassArray`` of shape ``shape`` filled with P1
        """
        return cls(np.full(shape, 0, dtype=np.int_))

    @classmethod
    def octave(cls, shape):
        """
        Same as unison() since octaves are equivalent to unisons in interval-class space.

        :param shape: the shape of the resulting array (tuple of integers)
        :return: a ``SpelledIntervalClassArray`` of shape ``shape`` filled with P1
        """
        return cls.unison(shape)

    @classmethod
    def chromatic_semitone(cls, shape):
        """
        Returns an array of the given shape filled with chromatic semitones (a1).

        :param shape: the shape of the resulting array (tuple of integers)
        :return: a ``SpelledIntervalClassArray`` of shape ``shape`` filled with a1
        """
        return cls(np.full(shape, 7, dtype=np.int_))

    def __add__(self, other):
        if type(other) == SpelledIntervalClass or type(other) == SpelledIntervalClassArray:
            return SpelledIntervalClassArray(self.fifths() + other.fifths())
        else:
            return NotImplemented

    def __sub__(self, other):
        if type(other) == SpelledIntervalClass or type(other) == SpelledIntervalClassArray:
            return SpelledIntervalClassArray(self.fifths() - other.fifths())
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, numbers.Integral) or\
           (hasattr(other, 'dtype') and issubclass(other.dtype.type, numbers.Integral)):
            return SpelledIntervalClassArray(self.fifths() * other)
        return NotImplemented
    
    def __neg__(self):
        return SpelledIntervalClassArray(-self.fifths())

    def __abs__(self):
        downs = self.direction() < 0
        abs_fifths = self.fifths().copy()
        # manually invert the intervals that point downwards
        abs_fifths[downs] = -abs_fifths[downs]
        return SpelledIntervalClassArray(abs_fifths)

    def direction(self):
        """
        Returns the (element-wise) directions of the intervals.
        The direction of each interval is determined by its shortest realization:
        ``m2``/``-M7`` is upward (1) while ``M7``/``-m2`` is downward (-1).
        Perfect unisons (``P1``) are neutral (0).
        
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        ds = self.diatonic_steps()
        mask = ds == 0
        out = np.ones_like(ds, dtype=np.int_)
        out[ds > 3] = -1
        out[mask] = np.sign((self.fifths()[mask] + 1) // 7)
        return out

    def ic(self):
        return self

    def embed(self):
        return SpelledIntervalArray.from_independent(self.fifths(), np.zeros_like(self.fifths()))

    def is_step(self):
        return np.isin(self.degree(), [0,1,6])

    # spelled interface

    def name(self):
        def intervalclass_name(fifths):
            return Spelled.interval_class_from_fifths(fifths)
        return np.vectorize(intervalclass_name, otypes=[np.str_])(self.fifths())

    def compare(self, other):
        """
        Element-wise comparison between two spelled arrays.

        Returns 0 where the elements are equal,
        1 where the first element is greater,
        and -1 where the second element is greater.

        Spelled interval classes use line-of-fifth ordering,
        for example ``P4 < P1 < P5 < M2``.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).
        To test the overall equality of two spelled arrays,
        use :py:meth:`array_equal <SpelledIntervalClasssArray.array_equal>`

        :param other: another ``SpelledIntervalClassArray`` or ``SpelledIntervalClass``
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledIntervalClass) or isinstance(other, SpelledIntervalClassArray):
            return np.sign((self - other).fifths())
        else:
            raise TypeError(f"Cannot elements of {type(self)} to {type(other)}.")

    def fifths(self):
        return self._fifths

    def octaves(self):
        return 0

    def internal_octaves(self):
        return 0

    def generic(self):
        return self.degree()

    def diatonic_steps(self):
        return self.degree()

    def alteration(self):
        return (self.fifths() + 1) // 7

    def onehot(self, fifth_range, dtype=int):
        """
        Returns a one-hot encoding of the interval classes in fifths as the innermost dimension.
        The range of fifths is given by ``fifth_range`` as a tuple ``(lower, upper)``.
        The outer shape of the output tensor is identical to the shape of the original array,
        so the resulting shape is ``original_shape + (n_fifths,)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param dtype: dtype of the resulting array (default: ``int``)
        :return: a one-hot tensor (numpy array)
        """
        flow, fhigh = fifth_range
        f = self.fifths()
        if (f < flow).any() or (f > fhigh).any():
            raise ValueError(f"The interval {self} is outside the given fifth range {fifth_range}.")
        # translate fifths and octaves to 0-based indices (from [low,high] to [0,high-low])
        f = f - flow

        # compute one-hot indices
        inner_shape = f.shape
        # a tuple of indices:
        # the first dimensions encode the position in the original array,
        # the last dimension encodes the fifth position.
        indices = tuple(np.indices(inner_shape)) + (f,)

        # initialize one-hot tensor with 0s
        out = np.zeros(inner_shape + (fhigh-flow+1,), dtype=dtype)
        # set all elements picked by the indices to 1
        out[indices] = 1
        return out

class SpelledPitchArray(SpelledArray, AbstractSpelledArrayPitch, Pitch):
    """
    Represents an array of spelled pitches.
    """

    _print_name = "asp"

    # constructors
    
    def __init__(self, fifths, octaves):
        """
        Takes two numpy arrays,
        one for fifths and one for (internal/dependent) octaves,
        both as integers.

        :param fifths: the internal fifths of each pitch (numpy array of integers)
        :param octaves: the internal octaves of each pitch (numpy array of integers)
        """
        # assert fifths.dtype == np.int_
        # assert octaves.dtype == np.int_
        if fifths.shape != octaves.shape:
            raise ValueError(f"Cannot create SpelledPitchArray from arrays of different sizes ({fifths.shape} and {octaves.shape}).")
        self._fifths = fifths
        self._octaves = octaves

    @staticmethod
    def from_independent(fifths, octaves):
        """
        Create a pitch array from fifths and indenpendent octaves.
        The fifths indicate the names of the pitches
        while the octaves indicate their octave numbers.

        :param fifths:  the internal fifths of each pitch (numpy array of integers)
        :param octaves: the external/independent octaves of each pitch (numpy array of integers)
        :return: the corresponding pitch array
        """
        return SpelledPitchArray(fifths, octaves - (fifths * 4) // 7)

    @staticmethod
    def from_strings(strings):
        """
        Create a pitch array from an array of strings.

        :param strings: an array-like of pitch notation strings
        :return: the corresponding pitch array
        """
        # assert isinstance(strings.dtype, np.str_)
        def parse_pitch(string):
            octave, fifth = Spelled.parse_pitch(string)
            if octave is None:
                raise ValueError(f"Missing octave specifier in pitch '{string}'.")
            return octave, fifth
        octaves, fifths = np.vectorize(parse_pitch, otypes=[np.int_, np.int_])(strings)
        return SpelledPitchArray.from_independent(fifths, octaves)

    @staticmethod
    def from_array(pitches):
        """
        Create a pitch array from an array of pitches.

        :param intervals: an array-like of ``SpelledPitch``
        :return: the corresponding pitch array
        """
        def from_pitch(pitch):
            return pitch.fifths(), pitch.internal_octaves()
        fifths, octaves = np.vectorize(from_pitch, otypes=[np.int_, np.int_])(pitches)
        return SpelledPitchArray(fifths, octaves)

    @staticmethod
    def from_onehot(onehot, fifth_low, octave_low):
        """
        Create a spelled pitch array from a one-hot tensor.
        ``fifth_low`` denotes the lower bound of the fifth range used in the vector,
        ``octave_low`` the lower bound of the octave range.
        The shape of the resulting array will be equivalent to the first n-2 dimensions of the input tensor.

        :param onehot: a one-hot tensor representing the pitches (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot tensor (integer)
        :param octave_low: the lowest octave expressible in the one-hot tensor (integer)
        :return: the corresponding pitch array
        """
        if (onehot.sum((-2,-1)) != 1).any():
            raise ValueError(f"{onehot} is not a one-hot tensor.")
        
        ones = np.where(onehot==1)
        indices = ones[:-2] # first n-2 dimensions are indices
        fifths_values = ones[-2] + fifth_low
        octs_values = ones[-1] + octave_low

        new_shape = onehot.shape[:-2]
        new_fifths = np.zeros(new_shape, dtype=int)
        new_fifths[indices] = fifths_values
        new_octaves = np.zeros(new_shape, dtype=int)
        new_octaves[indices] = octs_values
        return SpelledPitchArray.from_independent(new_fifths, new_octaves)

    # collection interface

    def __copy__(self):
        return SpelledPitchArray(copy.copy(self.fifths()), copy.copy(self.internal_octaves()))

    def __deepcopy__(self, memo):
        return SpelledPitchArray(copy.deepcopy(self.fifths(), memo), copy.deepcopy(self.internal_octaves(), memo))
    
    def __getitem__(self, index):
        f = self._fifths[index]
        o = self._octaves[index]
        if isinstance(f, numbers.Integral):
            return SpelledPitch.from_fifths_and_octaves(f, o)
        else:
            return SpelledPitchArray(f, o)

    def __setitem__(self, index, item):
        if isinstance(item, SpelledPitch) or isinstance(item, SpelledPitchArray):
            self._fifths[index] = item.fifths()
            self._octaves[index] = item.internal_octaves()
        else:
            raise TypeError(f"Cannot set elements of SpelledPitchArray to {type(item)}.")

    def __contains__(self, item):
        if isinstance(item, SpelledPitch):
            return ((self.fifths() == item.fifths()) & \
                    (self.internal_octaves() == item.internal_octaves())).any()
        else:
            return False

    def __len__(self):
        return len(self.fifths())
    
    class SpelledPitchArrayIter:
        def __init__(self, array):
            self._fifths_iter = array.fifths().__iter__()
            self._octaves_iter = array.internal_octaves().__iter__()

        def __next__(self):
            f = self._fifths_iter.__next__()
            o = self._octaves_iter.__next__()
            if isinstance(f, numbers.Integral):
                return SpelledPitch.from_fifths_and_octaves(f, o)
            else:
                return SpelledPitchArray(f, o)
        
    def __iter__(self):
        return self.SpelledPitchArrayIter(self)
        
    # Pitch interface

    def __add__(self, other):
        if type(other) == SpelledInterval or type(other) == SpelledIntervalArray:
            return SpelledPitchArray(self.fifths() + other.fifths(),
                                     self.internal_octaves() + other.internal_octaves())
        return NotImplemented

    def interval_from(self, other):
        if type(other) == SpelledPitch or type(other) == SpelledPitchArray:
            return SpelledIntervalArray(self.fifths() - other.fifths(),
                                     self.internal_octaves() - other.internal_octaves())
        else:
            raise TypeError(f"Cannot take interval between SpelledPitchArray and {type(other)}.")

    def pc(self):
        return SpelledPitchClassArray(self.fifths())

    def embed(self):
        return self

    # Spelled interface

    def compare(self, other):
        """
        Element-wise comparison between two spelled arrays.

        Returns 0 where the elements are equal,
        1 where the first element is greater,
        and -1 where the second element is greater.

        Spelled pitches use diatonic ordering,
        for example ``F4 < F#4 < F####4 < Gb4 < G4``.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).
        To test the overall equality of two spelled arrays,
        use :py:meth:`array_equal <SpelledPitch.array_equal>`

        :param other: another ``SpelledPitchArray`` or ``SpelledPitch``
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledPitch) or isinstance(other, SpelledPitchArray):
            return (self - other).direction()
        else:
            raise TypeError(f"Cannot elements of {type(self)} to {type(other)}.")

    def name(self):
        def pitch_name(fifths, octave):
            return f"{Spelled.pitch_class_from_fifths(fifths)}{octave}"
        return np.vectorize(pitch_name, otypes=[np.str_])(self.fifths(), self.octaves())

    def fifths(self):
        return self._fifths

    def octaves(self):
        return self.internal_octaves() + (self.fifths() * 4) // 7

    def internal_octaves(self):
        return self._octaves

    def alteration(self):
        return (self.fifths() + 1) // 7

    def letter(self):
        return ((self.degree() + 2) % 7 + ord('A')).astype(np.uint8).view('c').astype(np.str_)

    def onehot(self, fifth_range, octave_range, dtype=int):
        """
        Returns a one-hot encoding of the pitches in fifths and independent octaves as the innermost dimensions.
        The range of fifths and octaves is given by ``fifth_range`` and ``octave_range`` respectively,
        where each is a tuple ``(lower, upper)``.
        The outer shape of the output tensor is identical to the shape of the original array,
        so the resulting shape is ``original_shape + (n_fifths, n_octaves)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param octave_range: the (inclusive) range of octaves (pair of integers)
        :param dtype: dtype of the resulting array (default: ``int``)
        :return: a one-hot tensor (numpy array)
        """
        flow, fhigh = fifth_range
        olow, ohigh = octave_range
        f = self.fifths()
        o = self.octaves()
        if (f < flow).any() or (f > fhigh).any():
            raise ValueError(f"The pitch {self} is outside the given fifth range {fifth_range}.")
        if (o < olow).any() or (o > ohigh).any():
            raise ValueError(f"The pitch {self} is outside the given octave range {octave_range}.")
        # translate fifths and octaves to 0-based indices (from [low,high] to [0,high-low])
        f = f - flow
        o = o - olow

        # compute one-hot indices
        inner_shape = f.shape
        # a tuple of indices:
        # the first dimensions encode the position in the original array,
        # the last two dimensions encode the fifth and octave position.
        indices = tuple(np.indices(inner_shape)) + (f, o)

        # initialize one-hot tensor with 0s
        out = np.zeros(inner_shape + (fhigh-flow+1, ohigh-olow+1), dtype=dtype)
        # set all elements picked by the indices to 1
        out[indices] = 1
        return out

class SpelledPitchClassArray(SpelledArray, AbstractSpelledArrayPitch, Pitch):
    """
    Represents a spelled pitch class, i.e. a pitch without octave information.
    """

    _print_name = "aspc"

    def __init__(self, fifths):
        """
        Takes a numpy array of fifths as integers.

        :param fifths: the internal fifths of each pitch class (numpy array of integers)
        """
        self._fifths = fifths

    @staticmethod
    def from_strings(strings):
        """
        Create a pitch-class array from an array of strings.

        :param strings: an array-like of pitch-class notation strings
        :return: the corresponding pitch-class array
        """
        def parse_pc(string):
            octave, fifth = Spelled.parse_pitch(string)
            if octave is not None:
                raise ValueError(f"Pitch classes cannot have octave specifiers ({string}).")
            return fifth
        fifths = np.vectorize(parse_pc, otypes=[np.int_])(strings)
        return SpelledPitchClassArray(fifths)

    @staticmethod
    def from_array(pitches):
        """
        Create an pitch class array from an array of pitch classes.

        :param intervals: an array-like of ``SpelledPitchClass``
        :return: the corresponding pitch-class array
        """
        fifths = np.vectorize(lambda i: i.fifths(), otypes=[np.int_])(pitches)
        return SpelledPitchClassArray(fifths)

    @staticmethod
    def from_onehot(onehot, fifth_low):
        """
        Create a spelled pitch-class array from a one-hot tensor.
        ``fifth_low`` denotes the lower bound of the fifth range used in the vector.
        The shape of the resulting array will be equivalent to the first n-1 dimensions of the input tensor.

        :param onehot: a one-hot tensor representing the pitch classes (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot tensor (integer)
        :return: the corresponding pitch-class array
        """
        if (onehot.sum(-1) != 1).any():
            raise ValueError(f"{onehot} is not a one-hot tensor.")
        
        ones = np.where(onehot==1)
        indices = ones[:-1] # first n-1 dimensions are indices
        fifths_values = ones[-1] + fifth_low

        new_shape = onehot.shape[:-1]
        new_fifths = np.zeros(new_shape, dtype=int)
        new_fifths[indices] = fifths_values
        return SpelledPitchClassArray(new_fifths)

    # collection interface

    def __copy__(self):
        return SpelledPitchClassArray(copy.copy(self.fifths()))

    def __deepcopy__(self, memo):
        return SpelledPitchClassArray(copy.deepcopy(self.fifths(), memo))

    def __getitem__(self, index):
        f = self._fifths[index]
        if isinstance(f, numbers.Integral):
            return SpelledPitchClass.from_fifths(f)
        else:
            return SpelledPitchClassArray(f)

    def __setitem__(self, index, item):
        if isinstance(item, SpelledPitchClass) or isinstance(item, SpelledPitchClassArray):
            self._fifths[index] = item.fifths()
        else:
            raise TypeError(f"Cannot set elements of SpelledPitchClassArray to {type(item)}.")

    def __contains__(self, item):
        if isinstance(item, SpelledPitchClass):
            return item.fifths() in self.fifths()
        else:
            return False

    def __len__(self):
        return len(self.fifths())
    
    class SpelledPitchClassArrayIter:
        def __init__(self, array):
            self._fifths_iter = array.fifths().__iter__()
            
        def __next__(self):
            f = self._fifths_iter.__next__()
            if isinstance(f, numbers.Integral):
                return SpelledPitchClass.from_fifths(f)
            else:
                return SpelledPitchClassArray(f)
        
    def __iter__(self):
        return self.SpelledPitchClassArrayIter(self)
        
    # pitch interface

    def __add__(self, other):
        if type(other) == SpelledIntervalClass or type(other) == SpelledIntervalClassArray:
            return SpelledPitchClassArray(self.fifths() + other.fifths())
        else:
            return NotImplemented

    def interval_from(self, other):
        if type(other) == SpelledPitchClass or type(other) == SpelledPitchClassArray:
            return SpelledIntervalClassArray(self.fifths() - other.fifths())
        else:
            raise TypeError(f"Cannot take interval between SpelledPitchClassArray and {type(other)}.")

    def pc(self):
        return self

    def embed(self):
        return SpelledPitchArray.from_independent(self.fifths(), np.zeros_like(self.fifths()))

    # spelled interface

    def name(self):
        def pitchclass_name(fifths):
            return Spelled.pitch_class_from_fifths(fifths)
        return np.vectorize(pitchclass_name, otypes=[np.str_])(self.fifths())

    def compare(self, other):
        """
        Element-wise comparison between two spelled arrays.

        Returns 0 where the elements are equal,
        1 where the first element is greater,
        and -1 where the second element is greater.

        Spelled pitch classes use line-of-fifth ordering,
        for example ``Bb < F < C < G < D``.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).
        To test the overall equality of two spelled arrays,
        use :py:meth:`array_equal <SpelledPitchClassArray.array_equal>`

        :param other: another ``SpelledPitchClassArray`` or ``SpelledPitchClass``
        :return: an array of ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledPitchClass) or isinstance(other, SpelledPitchClassArray):
            return np.sign((self - other).fifths())
        else:
            raise TypeError(f"Cannot elements of {type(self)} to {type(other)}.")

    def fifths(self):
        return self._fifths

    def octaves(self):
        return 0

    def internal_octaves(self):
        return 0

    def alteration(self):
        return (self.fifths() + 1) // 7

    def letter(self):
        return ((self.degree() + 2) % 7 + ord('A')).astype(np.uint8).view('c').astype(np.str_)

    def onehot(self, fifth_range, dtype=int):
        """
        Returns a one-hot encoding of the pitch classes in fifths as the innermost dimension.
        The range of fifths is given by ``fifth_range`` as a tuple ``(lower, upper)``.
        The outer shape of the output tensor is identical to the shape of the original array,
        so the resulting shape is ``original_shape + (n_fifths,)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param dtype: dtype of the resulting array (default: ``int``)
        :return: a one-hot tensor (numpy array)
        """
        flow, fhigh = fifth_range
        f = self.fifths()
        if (f < flow).any() or (f > fhigh).any():
            raise ValueError(f"The interval {self} is outside the given fifth range {fifth_range}.")
        # translate fifths and octaves to 0-based indices (from [low,high] to [0,high-low])
        f = f - flow

        # compute one-hot indices
        inner_shape = f.shape
        # a tuple of indices:
        # the first dimensions encode the position in the original array,
        # the last dimension encodes the fifth position.
        indices = tuple(np.indices(inner_shape)) + (f,)

        # initialize one-hot tensor with 0s
        out = np.zeros(inner_shape + (fhigh-flow+1,), dtype=dtype)
        # set all elements picked by the indices to 1
        out[indices] = 1
        return out

# shorthand constructors

def asi(things, things2=None):
    """
    A quick way to construct a spelled-interval array.
    Takes an array-like of strings or spelled intervals,
    or two array-likes of integers (fifths and internal/dependent octaves).

    :param things: an array-like of strings / fifths (integers) / ``SpelledInterval``
    :param things2: an array-like of dependent octaves (integers), when providing fifths in the first parameter
    :return: a spelled-interval array of the same shape as the input
    """
    input = np.array(things)
    if input.dtype.type is np.str_ or input.dtype.type is np.string_:
        return SpelledIntervalArray.from_strings(input)
    if isinstance(input.flat[0], SpelledInterval):
        return SpelledIntervalArray.from_array(input)
    else:
        return SpelledIntervalArray(input, np.array(things2))

def asic(things):
    """
    A quick way to construct a spelled-interval-class array.
    Takes either an array-like of strings or spelled interval classes,
    or an array-like of integers (fifths).

    :param things: an array-like of strings / fifths (integers) / ``SpelledIntervalClass``
    :return: a spelled-interval-class array of the same shape as the input
    """
    input = np.array(things)
    if input.dtype.type is np.str_ or input.dtype.type is np.string_:
        return SpelledIntervalClassArray.from_strings(input)
    if isinstance(input.flat[0], SpelledIntervalClass):
        return SpelledIntervalClassArray.from_array(input)
    else:
        return SpelledIntervalClassArray(input)

def asp(things, things2=None):
    """
    A quick way to construct a spelled-pitch array.
    Takes either an array-like of strings or spelled pitches,
    or two array-likes of integers (fifths and internal/dependent octaves).

    :param things: an array-like of strings / fifths (integers) / ``SpelledPitch``
    :param things2: an array-like of dependent octaves (integers), when providing fifths in the first parameter
    :return: a spelled-pitch array of the same shape as the input
    """
    input = np.array(things)
    if input.dtype.type is np.str_ or input.dtype.type is np.string_:
        return SpelledPitchArray.from_strings(input)
    if isinstance(input.flat[0], SpelledPitch):
        return SpelledPitchArray.from_array(input)
    else:
        return SpelledPitchArray(input, np.array(things2))

def aspc(things):
    """
    A quick way to construct a spelled-pitch-class array.
    Takes either an array-like of strings or spelled pitch classes,
    or an array-like of integers (fifths).

    :param things: an array-like of strings / fifths (integers) / ``SpelledPitchClass``
    :return: a spelled-pitch-class array of the same shape as the input
    """
    input = np.array(things)
    if input.dtype.type is np.str_ or input.dtype.type is np.string_:
        return SpelledPitchClassArray.from_strings(input)
    if isinstance(input.flat[0], SpelledPitchClass):
        return SpelledPitchClassArray.from_array(input)
    else:
        return SpelledPitchClassArray(input)



#  Copyright (c) 2021 Robert Lieck

import numbers

from pitchtypes.basetypes import AbstractBase
from pitchtypes.spelled import Spelled
from pitchtypes.logfreq import LogFreq


class Enharmonic(AbstractBase):

    # how should Pitch and PitchClass types be printed
    _print_as_int = False
    _print_flat_sharp = 'sharp'

    @classmethod
    def print_options(cls, as_int=None, flat_sharp=None):
        if cls == Enharmonic:
            Enharmonic.Pitch.print_options(as_int=as_int, flat_sharp=flat_sharp)
            Enharmonic.PitchClass.print_options(as_int=as_int, flat_sharp=flat_sharp)
        if as_int is not None:
            cls._print_as_int = as_int
        if flat_sharp is not None:
            if flat_sharp not in ['sharp', 'flat']:
                raise ValueError("'flat_sharp' has to be one of ['sharp', 'flat']")
            else:
                cls._print_flat_sharp = flat_sharp
        if as_int is None and flat_sharp is None:
            print(f"print options in {cls.__name__}:\n"
                  f"    as_int: {cls._print_as_int}\n"
                  f"    flat_sharp: {cls._print_flat_sharp}")

    @staticmethod
    def pitch_class_name_from_midi(midi_pitch, flat_sharp):
        """
        Return the pitch class name for the given pitch in MIDI integer.
        :param midi_pitch: MIDI pitch
        :param flat_sharp: whether to use flats or sharps for accidentals
        :return: pitch class
        """
        if flat_sharp == "sharp":
            base_names = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
        elif flat_sharp == "flat":
            base_names = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"]
        else:
            raise ValueError("parameter 'flat_sharp' must be one of ['sharp', 'flat']")
        return base_names[midi_pitch % 12]

    def __init__(self, value, is_pitch, is_class, **kwargs):
        # pre-process value
        if isinstance(value, str):
            if is_pitch:
                if is_class:
                    value = Spelled.PitchClass(value=value).convert_to(EnharmonicPitchClass).value
                else:
                    value = Spelled.Pitch(value=value).convert_to(EnharmonicPitch).value
            else:
                if is_class:
                    value = Spelled.IntervalClass(value=value).convert_to(EnharmonicIntervalClass).value
                else:
                    value = Spelled.Interval(value=value).convert_to(EnharmonicInterval).value
        elif isinstance(value, numbers.Number):
            int_value = int(value)
            if int_value != value:
                raise ValueError(f"Expected integer pitch value but got {value}")
            value = int_value
            if is_class:
                value = value % 12
        # hand on initialisation to other base classes
        super().__init__(value=value, is_pitch=is_pitch, is_class=is_class, **kwargs)

    def convert_to_logfreq(self):
        raise NotImplementedError

    def __int__(self):
        return self.value

    def __repr__(self):
        return self.name()

    def name(self, *args, **kwargs):
        raise NotImplementedError


@Enharmonic.link_pitch_type()
class EnharmonicPitch(Enharmonic):

    def to_class(self):
        return self.PitchClass(value=self.value % 12)

    def name(self, as_int=None, flat_sharp=None):
        if as_int is None:
            as_int = self._print_as_int
        if flat_sharp is None:
            flat_sharp = self._print_flat_sharp
        if as_int:
            return str(self.value)
        return self.pitch_class_name_from_midi(self.value, flat_sharp=flat_sharp) + str(self.octaves())

    def octaves(self):
        return self.value // 12 - 1

    def freq(self):
        return 2 ** ((self.value - 69) / 12) * 440

    def convert_to_logfreq(self):
        return LogFreq.Pitch(self.freq(), is_freq=True)

    @property
    def midi(self):
        return self.value


@Enharmonic.link_interval_type()
class EnharmonicInterval(Enharmonic):
    def to_class(self):
        return self.IntervalClass(value=self.value % 12)

    def name(self):
        sign = "-" if self.value < 0 else ""
        return sign + str(abs(self.value))

    def octaves(self):
        return self.value // 12

    def convert_to_logfreq(self):
        return LogFreq.Interval(2 ** (self.value / 12), is_ratio=True)


@Enharmonic.link_pitch_class_type()
class EnharmonicPitchClass(Enharmonic):

    def name(self, as_int=None, flat_sharp=None):
        if as_int is None:
            as_int = self._print_as_int
        if flat_sharp is None:
            flat_sharp = self._print_flat_sharp
        if as_int:
            return str(self.value)
        return self.pitch_class_name_from_midi(self.value, flat_sharp=flat_sharp)

    def convert_to_logfreq(self):
        return LogFreq.PitchClass(2 ** ((self.value - 69) / 12) * 440, is_freq=True)


@Enharmonic.link_interval_class_type()
class EnharmonicIntervalClass(Enharmonic):
    def name(self):
        sign = "-" if self.value < 0 else ""
        return sign + str(abs(self.value))

    def convert_to_logfreq(self):
        return LogFreq.IntervalClass(2 ** (self.value / 12), is_ratio=True)



#  Copyright (c) 2021 Robert Lieck

from pitchtypes import Converters, Spelled, Enharmonic, LogFreq

def convert_spelled_to_enharmonic(spelled):
    if spelled.is_pitch:
        fifth_steps_from_f = spelled.fifths() + 1
        # get the base pitch in 0,...,11
        base_pitch = ((fifth_steps_from_f % 7 - 1) * 7) % 12
        # get the accidental, i.e. chromatic semitone steps to add to base pitch
        # (floor-divide (//) rounds down, for negative numbers that is equal to the remainder of division minus one)
        accidentals = fifth_steps_from_f // 7
        if spelled.is_class:
            return Enharmonic.PitchClass(value=base_pitch + accidentals)
        else:
            return Enharmonic.Pitch(value=12 * (spelled.octaves() + 1) + base_pitch + accidentals)
    else:
        # convert intervals by going via reference pitches
        if spelled.is_class:
            spelled_ref_point = Spelled.PitchClass("C")
            enharmonic_ref_point = Enharmonic.PitchClass("C")
        else:
            spelled_ref_point = Spelled.Pitch("C4")
            enharmonic_ref_point = Enharmonic.Pitch("C4")
        return enharmonic_ref_point - convert_spelled_to_enharmonic(spelled_ref_point - spelled)


Converters.register_converter(from_type=Spelled.Pitch,
                              to_type=Enharmonic.Pitch,
                              conv_func=convert_spelled_to_enharmonic)
Converters.register_converter(from_type=Spelled.Interval,
                              to_type=Enharmonic.Interval,
                              conv_func=convert_spelled_to_enharmonic)
Converters.register_converter(from_type=Spelled.PitchClass,
                              to_type=Enharmonic.PitchClass,
                              conv_func=convert_spelled_to_enharmonic)
Converters.register_converter(from_type=Spelled.IntervalClass,
                              to_type=Enharmonic.IntervalClass,
                              conv_func=convert_spelled_to_enharmonic)


Converters.register_converter(from_type=Enharmonic.Pitch,
                              to_type=LogFreq.Pitch,
                              conv_func=lambda enharmonic: enharmonic.convert_to_logfreq())
Converters.register_converter(from_type=Enharmonic.Interval,
                              to_type=LogFreq.Interval,
                              conv_func=lambda enharmonic: enharmonic.convert_to_logfreq())
Converters.register_converter(from_type=Enharmonic.PitchClass,
                              to_type=LogFreq.PitchClass,
                              conv_func=lambda enharmonic: enharmonic.convert_to_logfreq())
Converters.register_converter(from_type=Enharmonic.IntervalClass,
                              to_type=LogFreq.IntervalClass,
                              conv_func=lambda enharmonic: enharmonic.convert_to_logfreq())



#  Copyright (c) 2021 Robert Lieck

import numbers

import numpy as np

from pitchtypes import AbstractBase


class LogFreq(AbstractBase):
    """
    Represents a pitches and intervals in continuous frequency space with the frequency value stored in log
    representation.
    """

    _print_precision = 2

    @classmethod
    def print_precision(cls, precision=None):
        if precision is not None:
            if not isinstance(precision, numbers.Integral):
                raise ValueError(f"precision has to be an integer, got {precision}")
            if cls == LogFreq:
                for sub_cls in [LogFreq.Pitch, LogFreq.Interval, LogFreq.PitchClass, LogFreq.IntervalClass]:
                    sub_cls.print_precision(precision)
            cls._print_precision = precision
        return cls._print_precision

    @classmethod
    def _convert_freq_str(cls, s):
        if s.endswith("Hz"):
            return float(s[:-2])
        else:
            raise ValueError(f"String values have to be floats followed by 'Hz', but got '{s}'")

    def __init__(self, value, is_pitch, is_class, is_log=True, **kwargs):
        """
        Initialise from frequency or log-frequency value.
        :param value: frequency or log-frequency (default) value
        :param is_log: whether value is frequency or log-frequency
        """
        if is_log:
            value = float(value)
        else:
            value = np.log(value)
        super().__init__(value=value, is_pitch=is_pitch, is_class=is_class, **kwargs)

    def __float__(self):
        return float(self.value)


@LogFreq.link_pitch_type()
class LogFreqPitch(LogFreq):
    def __init__(self, value, is_freq=False, **kwargs):
        if isinstance(value, str):
            value = self._convert_freq_str(value)
            is_freq = True
        super().__init__(value, is_pitch=True, is_class=False, is_log=not is_freq, **kwargs)

    def __repr__(self):
        return f"{np.format_float_positional(self.freq(), fractional=True, precision=self._print_precision)}Hz"

    def to_class(self):
        return self.PitchClass(self.value, is_freq=False)

    def freq(self):
        return np.exp(self.value)


@LogFreq.link_interval_type()
class LogFreqInterval(LogFreq):
    def __init__(self, value, is_ratio=False, **kwargs):
        if isinstance(value, str):
            value = float(value)
            is_ratio = True
        super().__init__(value, is_pitch=False, is_class=False, is_log=not is_ratio, **kwargs)

    def __repr__(self):
        return f"{np.format_float_positional(self.ratio(), fractional=True, precision=self._print_precision)}"

    def to_class(self):
        return self.IntervalClass(self.value, is_ratio=False)

    def ratio(self):
        return np.exp(self.value)


@LogFreq.link_pitch_class_type()
class LogFreqPitchClass(LogFreq):
    def __init__(self, value, is_freq=False, **kwargs):
        if isinstance(value, str):
            value = self._convert_freq_str(value)
            is_freq = True
        if is_freq:
            value = np.log(value)
        else:
            value = float(value)
        value %= np.log(2)
        super().__init__(value, is_pitch=True, is_class=True, is_log=True, **kwargs)

    def __repr__(self):
        return f"{np.format_float_positional(self.freq(), fractional=True, precision=self._print_precision)}Hz"

    def freq(self):
        return np.exp(self.value)


@LogFreq.link_interval_class_type()
class LogFreqIntervalClass(LogFreq):
    def __init__(self, value, is_ratio=False, **kwargs):
        if isinstance(value, str):
            value = float(value)
            is_ratio = True
        if is_ratio:
            value = np.log(value)
        else:
            value = float(value)
        value %= np.log(2)
        super().__init__(value, is_pitch=False, is_class=True, is_log=True, **kwargs)

    def __repr__(self):
        return f"{np.format_float_positional(self.ratio(), fractional=True, precision=self._print_precision)}"

    def ratio(self):
        return np.exp(self.value)



#  Copyright (c) 2020 Robert Lieck
from typing import Iterable, Union, Any, Callable, Optional

import numpy as np
import abc

class AbstractBase:
    """
    This is the abstract base class for all pitch and interval types of the library.
    It provides some shared functionality and properties. See :doc:`/types/abstractbase` for
    more detailed explanations.
    """

    @staticmethod
    def set_func_attr(sub_type: Any,
                      flags: Iterable[Union[bool,None]],
                      names: Iterable[str],
                      funcs: Iterable[Callable]):
        """
        Add functions ``funcs`` as methods with ``names`` to class ``sub_type``, controlled by ``flags``.
        This is used by the decorators
        :meth:`link_pitch_type`, :meth:`link_interval_type`,
        :meth:`link_pitch_class_type` and :meth:`link_interval_class_type`.

        :param sub_type: class to add the methods to
        :param flags: Iterable of flags that control whether a particular method is added or not:
         if `True`, add the method, even if it already exists (i.e. overwrite existing methods);
         if `False`, don't add the method, even if it does not exist;
         if `None`, add the method if does not exist, otherwise do not add it
         (i.e. try to add but don't overwrite existing methods)
        :param names: name of the methods; if added, they will be accessible as via `*.name`
        :param funcs: callables (i.e. implementations of the methods)

        :meta private:
        """
        for flag, name, func in zip(flags, names, funcs):
            if flag or flag is None and name not in vars(sub_type):
                setattr(sub_type, name, func)

    @staticmethod
    def name_check(cls: Any, sub_type: Any, suffix: str, skip_name_check: bool):
        """
        Check if ``sub_type`` follows the standard naming convention. The `Pitch`, `Interval`, `PitchClass` and
        `IntervalClass` sub-type of a `Basetype` should be called `BasetypePitch`, `BasetypeInterval`,
        `BasetypePitchClass` and `BasetypeIntervalClass`. This is used by the decorators
        :meth:`link_pitch_type`, :meth:`link_interval_type`,
        :meth:`link_pitch_class_type` and :meth:`link_interval_class_type`.

        :param cls: base type
        :param sub_type: sub-type
        :param suffix: expected suffix (`Pitch`, `Interval`, `PitchClass`, or `IntervalClass`)
        :param skip_name_check: skip the check and don't raise
        :raises TypeError: if `sub_type.__name__` does not match `cls.__name__ + suffix`

        :meta private:
        """
        if not skip_name_check:
            got_name = sub_type.__name__
            expected_name = cls.__name__ + suffix
            if got_name != expected_name:
                raise TypeError(f"Got class named {got_name}, but expected {expected_name}. "
                                f"Use skip_name_check=True to suppress.")

    @classmethod
    def link_pitch_type(cls,
                        skip_name_check: bool = False,
                        create_init: Optional[bool] = None,
                        create_add: Optional[bool] = None,
                        create_sub: Optional[bool] = None,
                        create_to_class: Optional[bool] = None):
        """
        A decorator to link a pitch type to its base type.

        :meta private:
        """
        def decorator(sub_type):
            # link types
            cls.Pitch = sub_type
            sub_type._base_type = cls

            # create default functions
            def __init__(self, value, **kwargs):
                super(sub_type, self).__init__(value=value, is_pitch=True, is_class=False, **kwargs)

            def __add__(self, other):
                if type(other) == self.Interval:
                    return self.Pitch(self.value + other.value)
                return NotImplemented

            def __sub__(self, other):
                if type(other) == self.Pitch:
                    return self.Interval(self.value - other.value)
                elif type(other) == self.Interval:
                    return self.Pitch(self.value - other.value)
                return NotImplemented

            def to_class(self):
                return self.PitchClass(self.value)

            # set default functions
            AbstractBase.set_func_attr(sub_type,
                                       [create_init, create_add, create_sub, create_to_class],
                                       ['__init__', '__add__', '__sub__', 'to_class'],
                                       [__init__, __add__, __sub__, to_class])

            # check name
            AbstractBase.name_check(cls, sub_type, "Pitch", skip_name_check)

            return sub_type
        return decorator

    @classmethod
    def link_interval_type(cls,
                           skip_name_check: bool = False,
                           create_init: Optional[bool] = None,
                           create_add: Optional[bool] = None,
                           create_sub: Optional[bool] = None,
                           create_mul: Optional[bool] = None,
                           create_div: Optional[bool] = None,
                           create_neg: Optional[bool] = None,
                           create_to_class: Optional[bool] = None):
        """
        A decorator to link an interval type to its base type.

        :meta private:
        """
        def decorator(sub_type):
            # link types
            cls.Interval = sub_type
            sub_type._base_type = cls

            # create default functions
            def __init__(self, value, **kwargs):
                super(sub_type, self).__init__(value=value, is_pitch=False, is_class=False, **kwargs)

            def __add__(self, other):
                if type(other) == self.Interval:
                    return self.Interval(self.value + other.value)
                return NotImplemented

            def __sub__(self, other):
                if type(other) == self.Interval:
                    return self.Interval(self.value - other.value)
                return NotImplemented

            def __mul__(self, other):
                return sub_type(self.value * other)

            def __rmul__(self, other):
                return self.__mul__(other)

            def __truediv__(self, other):
                return self.__mul__(1 / other)

            def __neg__(self):
                return -1 * self

            def to_class(self):
                return self.IntervalClass(self.value)

            # set default functions
            AbstractBase.set_func_attr(
                sub_type,
                [create_init, create_add, create_sub, create_mul, create_mul, create_div, create_neg, create_to_class],
                ['__init__', '__add__', '__sub__', '__mul__', '__rmul__', '__truediv__', '__neg__', 'to_class'],
                [__init__, __add__, __sub__, __mul__, __rmul__, __truediv__, __neg__, to_class]
            )

            # perform name check
            AbstractBase.name_check(cls, sub_type, "Interval", skip_name_check)

            return sub_type
        return decorator

    @classmethod
    def link_pitch_class_type(cls,
                              skip_name_check: bool = False,
                              create_init: Optional[bool] = None,
                              create_add: Optional[bool] = None,
                              create_sub: Optional[bool] = None):
        """
        A decorator to link a pitch class type to its base type.

        :meta private:
        """
        def decorator(sub_type):
            # link types
            cls.PitchClass = sub_type
            sub_type._base_type = cls

            # create default functions
            def __init__(self, value, **kwargs):
                super(sub_type, self).__init__(value=value, is_pitch=True, is_class=True, **kwargs)

            def __add__(self, other):
                if type(other) == self.IntervalClass:
                    return self.PitchClass(self.value + other.value)
                return NotImplemented

            def __sub__(self, other):
                if type(other) == self.PitchClass:
                    return self.IntervalClass(self.value - other.value)
                elif type(other) == self.IntervalClass:
                    return self.PitchClass(self.value - other.value)
                return NotImplemented

            # set default functions
            AbstractBase.set_func_attr(sub_type,
                                       [create_init, create_add, create_sub],
                                       ['__init__', '__add__', '__sub__'],
                                       [__init__, __add__, __sub__])

            # check name
            AbstractBase.name_check(cls, sub_type, "PitchClass", skip_name_check)

            return sub_type
        return decorator

    @classmethod
    def link_interval_class_type(cls,
                                 skip_name_check: bool = False,
                                 create_init: Optional[bool] = None,
                                 create_add: Optional[bool] = None,
                                 create_sub: Optional[bool] = None,
                                 create_mul: Optional[bool] = None,
                                 create_div: Optional[bool] = None,
                                 create_neg: Optional[bool] = None):
        """
        A decorator to link an interval class type to its base type.

        :meta private:
        """
        def decorator(sub_type):
            # link types
            cls.IntervalClass = sub_type
            sub_type._base_type = cls

            # create default functions
            def __init__(self, value, **kwargs):
                super(sub_type, self).__init__(value=value, is_pitch=False, is_class=True, **kwargs)

            def __add__(self, other):
                if type(other) == self.IntervalClass:
                    return self.IntervalClass(self.value + other.value)
                return NotImplemented

            def __sub__(self, other):
                if type(other) == self.IntervalClass:
                    return self.IntervalClass(self.value - other.value)
                return NotImplemented

            def __mul__(self, other):
                return self.IntervalClass(self.value * other)

            def __rmul__(self, other):
                return self.__mul__(other)

            def __truediv__(self, other):
                return self.__mul__(1 / other)

            def __neg__(self):
                return -1 * self

            # set default functions
            AbstractBase.set_func_attr(
                sub_type,
                [create_init, create_add, create_sub, create_mul, create_mul, create_div, create_neg],
                ['__init__', '__add__', '__sub__', '__mul__', '__rmul__', '__truediv__', '__neg__'],
                [__init__, __add__, __sub__, __mul__, __rmul__, __truediv__, __neg__]
            )

            # perform name check
            AbstractBase.name_check(cls, sub_type, "IntervalClass", skip_name_check)

            return sub_type
        return decorator

    @staticmethod
    def create_subtypes():
        def decorator(cls):
            # for the four Pitch/Interval(Class) types:
            # 1) create a class
            # 2) name it appropriately
            # 3) link to base class

            # Pitch
            class Pitch(cls):
                pass
            Pitch.__name__ = cls.__name__ + "Pitch"
            cls.link_pitch_type()(Pitch)

            # Interval
            class Interval(cls):
                pass
            Interval.__name__ = cls.__name__ + "Interval"
            cls.link_interval_type()(Interval)

            # PitchClass
            class PitchClass(cls):
                pass
            PitchClass.__name__ = cls.__name__ + "PitchClass"
            cls.link_pitch_class_type()(PitchClass)

            # IntervalClass
            class IntervalClass(cls):
                pass
            IntervalClass.__name__ = cls.__name__ + "IntervalClass"
            cls.link_interval_class_type()(IntervalClass)

            # return the class (which now has the linked sub-types)
            return cls
        return decorator

    def __init__(self, value, is_pitch, is_class, **kwargs):
        # call __init__ on super to be cooperative in multi-inheritance,
        # otherwise this should just call object.__init__ and **kwargs should be empty
        super().__init__(**kwargs)
        # the class is frozen if __isfrozen__ == True
        # __setattr__ will check for that and raise an error
        # to set __isfrozen__ for the first time, we need to bypass __setattr__ via super
        # otherwise it would try to check on the not (yet) existing attribute
        super().__setattr__('__isfrozen__', False)
        # we can now set attributes normally (better than dynamically setting, because IDE's can autocomplete)
        self.is_pitch = is_pitch
        self.is_interval = not is_pitch
        self.is_class = is_class
        self.value = value
        # we now freeze the class so attributes cannot be changed anymore
        self.__isfrozen__ = True

    def __setattr__(self, key, value):
        if self.__isfrozen__:
            raise AttributeError("Class is frozen, attributes cannot be set")
        else:
            super().__setattr__(key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    def __eq__(self, other):
        if type(other) == type(self):
            assert self.is_pitch == other.is_pitch
            assert self.is_class == other.is_class
            if isinstance(self.value, np.ndarray) or isinstance(other.value, np.ndarray):
                return np.array_equal(self.value, other.value)
            else:
                return self.value == other.value
        return False

    def __hash__(self):
        if isinstance(self.value, np.ndarray):
            assert self.value.flags.writeable is False
            return hash((self.__class__.__name__, self.value.data.tobytes(), self.is_pitch, self.is_class))
        else:
            return hash((self.__class__.__name__, self.value, self.is_pitch, self.is_class))

    def convert_to(self, other_type):
        return Converters.convert(self, other_type)

class Interval(abc.ABC):
    """
    The basic interface implemented by every interval (and interval class) type.
    """

    # fixed intervals

    @classmethod
    @abc.abstractmethod
    def unison(cls):
        """
        Create the unison interval of this type.

        :return: a unison interval
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def octave(cls):
        """
        Create the octave interval of this type.

        :return: an octave interval
        """
        raise NotImplementedError

    # operators

    # @abc.abstractmethod
    def __eq__(self, other):
        raise NotImplementedError

    # @abc.abstractmethod
    def __add__(self, other):
        """
        Returns the sum of two intervals.
        """
        raise NotImplementedError

    # @abc.abstractmethod
    def __sub__(self, other):
        """
        Returns the difference of two intervals.
        """
        raise NotImplementedError
    
    # @abc.abstractmethod
    def __mul__(self, other):
        """
        Returns an integer multiple of the interval.
        """
        raise NotImplementedError
    
    def __rmul__(self, other):
        """
        Returns an integer multiple of the interval.
        """
        return self.__mul__(other)
    
    # @abc.abstractmethod
    def __neg__(self):
        """
        Returns the inversion of the interval (same size in the other direction).
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __abs__(self):
        """
        For downward intervals, return their upward counterpart, otherwise just return the interval itself.
        """
        raise NotImplementedError

    
    # other interface methods
    
    @abc.abstractmethod
    def direction(self):
        """
        Determines the direction of the interval:
        1 for up, -1 for down and 0 for neutral.
        Different types may have different conventions for the direction of an interval.

        :return: the interval's direction (``-1`` / ``0`` / ``1``, integer)
        """
        raise NotImplementedError

    def abs(self):
        """
        For downward intervals, return their upward counterpart, otherwise just return the interval itself.

        :return: the absolute interval
        """
        return abs(self)
    
    @abc.abstractmethod
    def ic(self):
        """
        Return the interval class that corresponds to this interval.
        If the interval is already an interval class, it is returned itself.

        :return: the interval's corresponding interval class
        """
        raise NotImplementedError

    def to_class(self):
        """
        Alias for ic(), but also supported by pitch types.

        :return: the interval's corresponding interval class
        """
        return self.ic()

    @abc.abstractmethod
    def embed(self):
        """
        For interval classes, return an embedding into the interval space in a (type-dependent) default octave.
        For non-class intervals, return the interval itself.

        :return: a non-class version of this interval
        """
        raise NotImplementedError


class Chromatic(abc.ABC):
    """
    Some intervals have the notion of a chromatic semitone and implement this interface.
    """

    @classmethod
    @abc.abstractmethod
    def chromatic_semitone(cls):
        """
        Return a chromatic semitone (augmented unison) of this type.

        :return: a chromatic semitone interval
        """
        raise NotImplementedError


class Diatonic(abc.ABC):
    """
    Some intervals have a notion of a diatonic step and implement this interface.
    """

    @abc.abstractmethod
    def is_step(self):
        """
        Return True if the interval is considered a step, False otherwise.

        :return: the stepness of the interval (boolean)
        """
        raise NotImplementedError


class Pitch(abc.ABC):
    """
    The basic interface that is implemented by every pitch (and pitch class) type.
    """

    # arithmetic operations

    # @abc.abstractmethod
    def __eq__(self, other):
        raise NotImplementedError

    # @abc.abstractmethod
    def __add__(self, other):
        """
        Returns the pitch transposed by an interval

        :param other: an interval of matching type
        :return: the pitch transposed by ``other``
        """
        raise NotImplementedError

    def __sub__(self, other):
        """
        When subtracting another pitch (p1 - p2), return the interval from p2 to p1.
        When subtracting an interval (p - i), transpose the pitch by the inverse interval (p + -i).

        :param other: pitch or interval
        :return: if ``other`` is an interval, the transposed pitch;
                 if ``other`` is a pitch, the interval between both pitches
        """
        if isinstance(other, Pitch):
            try:
                return self.interval_from(other)
            except:
                return NotImplemented
        elif isinstance(other, Interval):
            return self + (-other)
        else:
            return NotImplemented

    # other pitch operations

    @abc.abstractmethod
    def interval_from(self, other):
        """
        Computes the interval from another pitch to this pitch.

        :param other: another pitch
        :return: the interval from other to self
        """
        raise NotImplementedError

    def interval_to(self, other):
        """
        Computes the interval from this pitch to another pitch.

        :param other: another pitch
        :return: the interval from self to other
        """
        return - self.interval_from(other)
    
    @abc.abstractmethod
    def pc(self):
        """
        Returns the pitch class corresponding to the pitch.
        For pitch classes, it returns the pitch class itself.

        :return: the pitch class that corresponds to this pitch
        """
        raise NotImplementedError

    def to_class(self):
        """
        Alias for pc(), but also supported by interval types.

        :return: the pitch class that corresponds to this pitch
        """
        return self.pc()

    @abc.abstractmethod
    def embed(self):
        """
        For a pitch class, returns the corresponding pitch in a (type-dependent) default octave.
        For non-class pitches, returns the pitch itself.

        :return: a non-class version of this pitch
        """
        raise NotImplementedError


class Converters:

    # store converters for classes derived from Pitch;
    # it's a dict of dicts, so that _converters[A][B] returns is a list of functions that, when executed
    # successively, converts A to B
    _converters = {}

    @staticmethod
    def convert(obj, to_type):
        # skip self-conversion
        if type(obj) == to_type:
            ret = obj
        else:
            # use conversion pipeline starting with the object itself
            ret = obj
            # sequentially apply converters from pipeline
            for converter in Converters.get_converter(type(obj), to_type):
                ret = converter(ret)
        # checks
        assert isinstance(ret, to_type), f"Conversion failed, expected type {to_type} but got {type(ret)}"
        assert obj.is_pitch == ret.is_pitch, f"{obj.is_pitch} {ret.is_pitch}"
        assert obj.is_class == ret.is_class, f"{obj.is_class} {ret.is_class}"
        return ret

    @staticmethod
    def get_converter(from_type, to_type=None):
        # return dedicated converter if other_type was specified or list of existing converters otherwise
        if to_type is not None:
            all_converters = Converters.get_converter(from_type)
            try:
                return all_converters[to_type]
            except KeyError:
                raise NotImplementedError(f"Type '{from_type}' does not have any converter registered for type "
                                          f"'{to_type}'")
        else:
            try:
                return Converters._converters[from_type]
            except KeyError:
                raise NotImplementedError(f"There are no converters registered for type '{from_type}'")

    @staticmethod
    def register_converter(from_type, to_type, conv_func,
                           overwrite_explicit_converters=False,
                           overwrite_implicit_converters=False,
                           create_implicit_converters=False):
        """
        Register a converter from from_type to other type. The converter function should be function taking as its
        single argument an from_type object and returning an other_type object.
        :param to_type: other type derived from AbstractBase, which the converter function converts to
        :param conv_func: converter function from from_type to other_type
        :param overwrite_explicit_converters: can be True, False, or None (default); if True and there exists an
        explicit converter (i.e. the list of converter functions is of length 1), replace it by this converter function;
        if False raise a ValueError if an explicit converter exists
        :param overwrite_implicit_converters: if there exists an implicit converter (i.e. the list of converter functions
        is of length greater than 1) replace it by this converter function
        :param create_implicit_converters: if there is an (explicit or implicit) converter from type X to type
        from_type, add an implicit converter from type X to other_type by extending the list of converter functions from
        X to from_type by this converter function; if there already exists an (explicit or implicit) converter from X to
        other_type, it will not be overwritten
        """
        # not self-conversion
        if from_type == to_type:
            raise TypeError(f"Not allowed to add converters from a type to itself (from: {from_type}, to: {to_type})")
        # initialise converter dict if it does not exist
        if from_type not in Converters._converters:
            Converters._converters[from_type] = {}
        # get existing converters from from_type to to_type and decide whether to set new converter
        set_new_converter = False
        try:
            converter = Converters.get_converter(from_type, to_type)
        except NotImplementedError:
            # no existing converters
            set_new_converter = True
        else:
            # implicit or explicit converter are already registered
            if len(converter) == 1:
                # explicit converter
                if overwrite_explicit_converters:
                    set_new_converter = True
                else:
                    raise ValueError("An explicit converter already exists. Set overwrite_explicit_converters=True to "
                                     "overwrite.")
            else:
                # implicit converter
                if overwrite_implicit_converters:
                    set_new_converter = True
                else:
                    raise ValueError("An implicit converter already exists. Set overwrite_implicit_converters=True to "
                                     "overwrite.")
        # set the new converter
        if set_new_converter:
            Converters._converters[from_type][to_type] = [conv_func]
        # extend implicit converters
        if create_implicit_converters:
            for another_from_type, other_converters in Converters._converters.items():
                # remember new converters to not change dict while iterating over it
                new_converters = []
                for another_to_type, converter_pipeline in other_converters.items():
                    # trying to prepend this converter (but don't add implicit self-converters)
                    # from_type --> another_to_type := (from_type --> to_type) + (another_from_type --> another_to_type)
                    if to_type == another_from_type and from_type != another_to_type:
                        # get existing converters from_type --> ???
                        converters = Converters.get_converter(from_type)
                        # add the extended converter if one does not exist
                        if another_to_type not in converters:
                            converters[another_to_type] = [conv_func] + converter_pipeline
                    # try to append this converter (but don't add implicit self-converters)
                    # another_from_type --> to_type := (another_from_type --> another_to_type) + ( from_type --> to_type)
                    if another_to_type == from_type and another_from_type != to_type:
                        # already initialised and we have the existing converters another_from_type --> ???
                        # add the extended converter if one does not exist
                        if to_type not in other_converters:
                            new_converters.append((to_type, converter_pipeline + [conv_func]))
                # insert new converters
                for another_to_type, converter_pipeline in new_converters:
                    other_converters[another_to_type] = converter_pipeline



#  Copyright (c) 2020 Robert Lieck

from .basetypes import *
from .spelled import *
from .spelled_array import *
from .enharmonic import *
from .logfreq import *
from .harmonic import *
from . import converters



#  Copyright (c) 2021 Robert Lieck

import numpy as np

from pitchtypes.basetypes import AbstractBase


class Harmonic(AbstractBase):

    @staticmethod
    def parse_exponents(exponents):
        if isinstance(exponents, str):
            # remove all whitespace
            exponents_ = "".join(exponents.split())
            # assert starts and ends with '[' and ']' respectively
            if not (exponents_.startswith("[") and exponents_.endswith("]")):
                raise ValueError(f"'exponents' has to start and end with '[' and ']', respectively")
            try:
                exponents = np.array(exponents_[1:-1].split(','), dtype=int)
            except ValueError as e:
                raise ValueError(f"Could not interpret {exponents} as array of integers: {e}")
        else:
            exponents = np.array(exponents, dtype=int)
        return exponents

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


@Harmonic.link_interval_type()
class HarmonicInterval(Harmonic):
    def __init__(self, exponents):
        super().__init__(value=self.parse_exponents(exponents=exponents),
                         is_pitch=False,
                         is_class=False)

    def __repr__(self):
        return f"{self.__class__.__name__}({list(self.value)})"

    def to_class(self):
        return self.IntervalClass(exponents=self.value[1:].copy())


@Harmonic.link_interval_class_type()
class HarmonicIntervalClass(Harmonic):
    def __init__(self, exponents):
        super().__init__(value=self.parse_exponents(exponents=exponents),
                         is_pitch=False,
                         is_class=True)

    def __repr__(self):
        return f"{self.__class__.__name__}({[None] + list(self.value)})"



#  Copyright (c) 2022 Robert Lieck, Christoph Finkensiep

import numbers
import re
import abc
import functools

import numpy as np

from pitchtypes.basetypes import AbstractBase, Pitch, Interval, Diatonic, Chromatic

@functools.total_ordering
class Spelled(AbstractBase):
    """
    A common base class for spelled pitch and interval types.
    See below for a set of common operations.
    """
    _pitch_regex = re.compile("^(?P<class>[A-G])(?P<modifiers>(b*)|(#*))(?P<octave>(-?[0-9]+)?)$")
    _interval_regex = re.compile("^(?P<sign>[-+])?("
                                 "(?P<quality0>P)(?P<generic0>[145])|"          # perfect intervals
                                 "(?P<quality1>|(M)|(m))(?P<generic1>[2367])|"  # imperfect intervals
                                 "(?P<quality2>(a+)|(d+))(?P<generic2>[1-7])"   # augmeted/diminished intervals
                                 ")(?P<octave>(:-?[0-9]+)?)$")

    @staticmethod
    def parse_pitch(s):
        """
        Parse a string as a spelled pitch or spelled pitch class. Returns a tuple (octave, fifths), where octave
        indicates the octave the pitch lies in (None for spelled pitch classes) and fifths indicates the steps taken
        along the line of fifths.
        
        :param s: string to parse
        :return: (octave, fifths)

        :meta private:
        """
        if not isinstance(s, str):
            raise TypeError(f"expected string as input, got {s}")
        # convert unicode flats and sharps (♭ -> b and ♯ -> #)
        s = s.replace("♭", "b")
        s = s.replace("♯", "#")
        # match with regex
        pitch_match = Spelled._pitch_regex.match(s)
        if pitch_match is None:
            raise ValueError(f"could not match '{s}' with regex: '{Spelled._pitch_regex.pattern}'")
        octave = pitch_match['octave']
        # initialise fifth steps from diatonic pitch class
        fifth_steps = Spelled.fifths_from_diatonic_pitch_class(pitch_match['class'])
        # add modifiers
        if "#" in pitch_match['modifiers']:
            fifth_steps += 7 * len(pitch_match['modifiers'])
        else:
            fifth_steps -= 7 * len(pitch_match['modifiers'])
        # add octave
        if octave == "":
            return None, fifth_steps
        else:
            return int(octave), fifth_steps

    @staticmethod
    def parse_interval(s):
        """
        Parse a string as a spelled interval or spelled interval class. Returns a tuple (sign, octave, fifths), where
        sign is +1 or -1 and indicates the sign given in the string (no sign means positive), octave indicates the
        number of full octave steps (in positive or negative direction; None for spelled interval classes), and fifths
        indicates the steps taken along the line of fifths (i.e. not actual fifth steps that would add to the octaves).
        
        :param s: string to parse
        :return: (sign, octave, fifths)

        :meta private:
        """
        if not isinstance(s, str):
            raise TypeError("expecte string as input, got {s}")
        interval_match = Spelled._interval_regex.match(s)
        if interval_match is None:
            raise ValueError(f"could not match '{s}' with regex: '{Spelled._interval_regex.pattern}'")
        # get quality and generic interval (first corresponding group that is not None)
        for i in range(3):
            g = interval_match[f"generic{i}"]
            q = interval_match[f"quality{i}"]
            if g is not None and q is not None:
                generic = int(g)
                quality = q
                break
        else:
            raise RuntimeError(f"Could not match generic interval and quality, this is a bug in the regex ("
                               f"{[interval_match[f'generic{i}'] for i in range(3)]}, "
                               f"{[interval_match[f'quality{i}'] for i in range(3)]}"
                               f")")
        # initialise value with generic interval classes
        fifth_steps = Spelled.fifths_from_generic_interval_class(generic)
        # add modifiers
        if quality in ["P", "M"]:
            pass
        elif quality == "m":
            fifth_steps -= 7
        elif "a" in quality:
            fifth_steps += 7 * len(quality)
        elif "d" in quality:
            if generic in [4, 1, 5]:
                fifth_steps -= 7 * len(quality)
            else:
                fifth_steps -= 7 * (len(quality) + 1)
        else:
            raise RuntimeError(f"Initialization from string failed: "
                               f"Unexpected interval quality '{quality}'. This is a bug and "
                               f"means that either the used regex is bad or the handling code.")
        # get octave
        if interval_match['octave'][1:] == "":
            octave = None
        else:
            octave = int(interval_match['octave'][1:])
        # get sign and bring adapt fifth steps
        if interval_match['sign'] == '-':
            sign = -1
        else:
            sign = 1
        return sign, octave, fifth_steps

    @staticmethod
    def pitch_class_from_fifths(fifth_steps):
        """
        Return the pitch class given the number of steps along the line of fifths

        :param fifth_steps: number of steps along the line of fifths
        :return: pitch class (e.g. C, Bb, F##, Abbb etc.)

        :meta private:
        """
        base_pitch = ["F", "C", "G", "D", "A", "E", "B"][(fifth_steps + 1) % 7]
        flat_sharp = (fifth_steps + 1) // 7
        return base_pitch + ('#' if flat_sharp > 0 else 'b') * abs(flat_sharp)

    @staticmethod
    def interval_quality_from_fifths(fifth_steps):
        """
        Return the interval quality (major, minor, perfect, augmented, diminished, doubly-augmented etc) given the
        number of steps along the line of fifths.
        
        :param fifth_steps: number of steps along the line of fifths
        :return: interval quality (M, m, p, a, d, aa, dd, aaa, ddd etc)

        :meta private:
        """
        if -5 <= fifth_steps <= 5:
            quality = ['m', 'm', 'm', 'm', 'P', 'P', 'P', 'M', 'M', 'M', 'M'][fifth_steps + 5]
        elif fifth_steps > 5:
            quality = 'a' * ((fifth_steps + 1) // 7)
        else:
            quality = 'd' * ((-fifth_steps + 1) // 7)
        return quality

    @staticmethod
    def diatonic_steps_from_fifths(fifth_steps):
        """
        Return the number of diatonic steps corresponding to the number of steps on the line of fifths
        (`4 * fifth_steps`).
        
        :param fifth_steps: number of fifth steps
        :return: number of diatonic steps

        :meta private:
        """
        return 4 * fifth_steps

    @staticmethod
    def generic_interval_class_from_fifths(fifth_steps):
        """
        Return the generic interval class corresponding to the given number of fifths. This corresponds to the number of
        diatonic steps plus one. The generic interval also corresponds to the scale degree when interpreted as the tone
        reached when starting from the tonic.
        
        :param fifth_steps: number of fifth steps
        :return: scale degree (integer in 1,...,7)

        :meta private:
        """
        return Spelled.diatonic_steps_from_fifths(fifth_steps) % 7 + 1

    @staticmethod
    def interval_class_from_fifths(fifths, inverse=False):
        """
        Return the interval class corresponding to the given number of steps along the line of fifths. This function
        combines Spelled.interval_quality_from_fifths and Spelled.generic_interval_class_from_fifths. Specifying
        inverse=True (default is False) returns the inverse interval class (m2 for M7, aa4 for dd5 etc.).
        
        :param fifths: number of fifth steps
        :param inverse: whether to return the inverse interval class
        :return: interval class (p1, M3, aa6 etc.)

        :meta private:
        """
        if inverse:
            fifths = -fifths
        return f"{Spelled.interval_quality_from_fifths(fifths)}" \
               f"{Spelled.generic_interval_class_from_fifths(fifths)}"

    @staticmethod
    def _degree_from_fifths_(fifths):
        """
        Return the scale degree of a pitch/interval based on its fifths.
        Helper function for degree()

        :meta private:
        """
        return (fifths*4) % 7

    @staticmethod
    def fifths_from_diatonic_pitch_class(pitch_class):
        """
        Return the number of steps along the line of fifths corresponding to a diatonic pitch class.
        
        :param pitch_class: a diatonic pitch class; character in A, B, C, D, E, F, G
        :return: fifth steps; an integer in -1, 0, ... 5

        :meta private:
        """
        pitch_classes = "ABCDEFG"
        if pitch_class not in pitch_classes:
            pitch_classes = "', '".join(pitch_classes)
            raise ValueError(f"diatonic pitch class must be one of '{pitch_classes}', but got {pitch_class}")
        return {"F": -1, "C": 0, "G": 1, "D": 2, "A": 3, "E": 4, "B": 5}[pitch_class]

    @staticmethod
    def fifths_from_generic_interval_class(generic):
        """
        Return the number of steps along the line of fifths corresponding to the given generic interval:
        (2 * generic - 1) % 7 - 1.
        
        :param generic: generic interval (integer in 1,...,7)
        :return: fifth steps (integer in -1, 0, ..., 5)

        :meta private:
        """
        if not isinstance(generic, numbers.Integral) or not (1 <= generic <= 7):
            raise ValueError(f"generic interval must be an integer between 1 and 7 (incl.), got {generic}")
        return (2 * generic - 1) % 7 - 1

    def __init__(self, value, is_pitch, is_class, **kwargs):
        super().__init__(value=value, is_pitch=is_pitch, is_class=is_class, **kwargs)
        if not is_class:
            self.value.flags.writeable = False

    def __repr__(self):
        return self.name()

    def name(self):
        """
        The name of the pitch or interval in string notation

        :return: the object's notation name (string)
        """
        raise NotImplementedError

    def compare(self, other):
        """
        Comparison between two spelled types.

        Returns 0 if the objects are equal,
        1 if the first object (``self``) is greater,
        and -1 if the second object (``other``) is greater.

        The respective ordering differs between types.
        Non-class pitches and intervals use diatonic ordering,
        interval/pitch classes use line-of-fifths ordering.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).

        :param other: an object to compare to (same type as ``self``)
        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        raise NotImplementedError

    def __eq__(self, other):
        try:
            return self.compare(other) == 0
        except TypeError:
            return NotImplemented

    def __lt__(self, other):
        try:
            return self.compare(other) == -1
        except TypeError:
            return NotImplemented
    
    # Spelled interface:

    def fifths(self):
        """
        Return the position of the interval on the line of fifths.

        :return: fifth position (integer)
        """
        raise NotImplementedError

    def octaves(self):
        """
        For intervals, return the number of octaves the interval spans.
        Negative intervals start with -1, decreasing.
        For pitches, return the absolute octave of the pitch.

        :return: external/independent octave (integer)
        """
        raise NotImplementedError

    def internal_octaves(self):
        """
        Return the internal octave representation of a pitch,
        which is dependent on the fifths.

        Only use this if you know what you are doing.

        :return: internal/dependent octave (integer)
        """
        raise NotImplementedError

    def degree(self):
        """
        Return the "relative scale degree" (0-6) to which the interval points
        (unison=0, 2nd=1, octave=0, 2nd down=6, etc.).
        For pitches, return the integer that corresponds to the letter (C=0, D=1, ...).

        :return: degree (integer)
        """
        return self._degree_from_fifths_(self.fifths())

    def alteration(self):
        """
        Return the number of semitones by which the interval is altered from its the perfect or major variant.
        Positive alteration always indicates augmentation,
        negative alteration indicates diminution (minor or smaller) of the interval.
        For pitches, return the accidentals (positive=sharps, negative=flats, 0=natural).

        :return: alteration (integer)
        """
        raise NotImplementedError

    def onehot(self):
        """
        Return a one-hot encoded tensor representing the object.
        Specialized versions of this method take ranges for their respective dimensions.

        :return: one-hot encoding of the object (numpy array)
        """
        raise NotImplementedError


class AbstractSpelledInterval(abc.ABC):
    """
    The interface for spelled interval types.
    """
    @abc.abstractmethod
    def generic(self):
        """
        Return the generic interval, i.e. the number of diatonic steps modulo octave.
        Unlike degree(), the result respects the sign of the interval
        (unison=0, 2nd up=1, 2nd down=-1).

        :return: generic interval (integer)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def diatonic_steps(self):
        """
        Return the diatonic steps of the interval (unison=0, 2nd=1, ..., octave=7, ...).
        Respects both direction and octaves.

        :return: number of diatonic steps (integer)
        """
        raise NotImplementedError

class AbstractSpelledPitch(abc.ABC):
    """
    The interface for spelled pitch types.
    """
    @abc.abstractmethod
    def letter(self):        
        """
        Returns the letter associated with the pitch (without accidentals).

        :return: letter (single-character string)
        """
        raise NotImplementedError

@Spelled.link_pitch_type()
class SpelledPitch(Spelled, AbstractSpelledPitch, Pitch):
    """
    Represents a spelled pitch.
    """
    def __init__(self, value):
        """        
        Takes a string consisting of the form
        ``<letter><accidentals?><octave>``, e.g. ``"C#4"``, ``"E5"``, or ``"Db-2"``.
        Accidentals may be written as ASCII symbols (#/b)
        or with unicode symbols (♯/♭), but not mixed within the same note.

        :param value: a string or internal numeric representation of the pitch
        """
        if isinstance(value, str):
            octaves, fifths = self.parse_pitch(value)
            assert isinstance(octaves, numbers.Integral)
            assert isinstance(fifths, numbers.Integral)
            # correct for octaves taken by fifth steps
            octaves -= Spelled.diatonic_steps_from_fifths(fifths) // 7
            value = np.array([octaves, fifths])
        else:
            octaves, fifths = value
            assert isinstance(octaves, numbers.Integral)
            assert isinstance(fifths, numbers.Integral)
            value = np.array([octaves, fifths])
        assert isinstance(fifths, numbers.Integral)
        assert isinstance(octaves, numbers.Integral)
        super().__init__(value=value, is_pitch=True, is_class=False)

    @staticmethod
    def from_fifths_and_octaves(fifths, octaves):
        """
        Create a pitch by directly providing its internal fifths and octaves.

        Each pitch is represented relative to C0
        by moving the specified number of fifths and octaves upwards
        (or downwards for negative values).

        :param fifths: the fifth (= pitch class) of the pitch (integer)
        :param octaves: the internal/dependent octave of the pitch (integer)
        :return: the resulting pitch (SpelledPitch)
        """
        return SpelledPitch((octaves, fifths))

    @staticmethod
    def from_independent(fifths, octaves):
        """
        Create a pitch from fifths (pitch class) and independent/external octaves (octave number).

        :param fifths: the fifth (= pitch class) of the pitch (integer)
        :param octaves: the external/independent octave of the pitch (integer)
        :return: the resulting pitch (SpelledPitch)
        """
        return SpelledPitch.from_fifths_and_octaves(fifths, octaves - (fifths * 4) // 7)

    @staticmethod
    def from_onehot(onehot, fifth_low, octave_low):
        """
        Create a spelled pitch from a one-hot matrix.

        Requires the lower bounds of the fifth and octave range used by the one-hot matrix.

        :param onehot: a one-hot matrix representing the pitch (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot matrix
        :param octave_low: the lowest octave expressible in the one-hot matrix
        :return: the resulting pitch (SpelledPitch)
        """
        if onehot.sum() != 1:
            raise ValueError(f"{onehot} is not a one-hot vector.")
        fs, os = np.where(onehot==1)
        return SpelledPitch.from_independent(fs[0] + fifth_low, os[0] + octave_low)
    
    # Pitch interface

    def interval_from(self, other):
        if type(other) == SpelledPitch:
            octaves1, fifths1 = self.value
            octaves2, fifths2 = other.value
            return SpelledInterval.from_fifths_and_octaves(fifths1-fifths2, octaves1-octaves2)
        else:
            raise TypeError(f"Cannot take interval between SpelledPitch and {type(other)}.")
    
    def to_class(self):
        return self.PitchClass(self.fifths())
    
    def pc(self):
        return self.to_class()

    def embed(self):
        return self

    # Spelled interface

    def name(self):
        return f"{self.pitch_class_from_fifths(self.fifths())}{self.octaves()}"

    def compare(self, other):
        """
        Comparison between two spelled pitches according to diatonic ordering.

        Returns 0 if the objects are equal,
        1 if the first pitch (``self``) is higher,
        and -1 if the second pitch (``other``) is higher.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).

        :param other: a pitch to compare to (SpelledPitch)
        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledPitch):
            return (self-other).direction()
        else:
            raise TypeError(f"Cannot compare {type(self)} with {type(other)}.")

    def fifths(self):
        return self.value[1]

    def octaves(self):
        return self.value[0] + self.diatonic_steps_from_fifths(self.fifths()) // 7

    def internal_octaves(self):
        return self.value[0]

    def alteration(self):
        return (self.fifths() + 1) // 7

    def letter(self):
        return chr(ord('A') + (self.degree() + 2) % 7)

    def onehot(self, fifth_range, octave_range, dtype=int):
        """
        Returns a one-hot encoding of the pitch in fifths (first dimension) and external octaves (second dimension).
        The range of fifths and octaves is given by ``fifth_range`` and ``octave_range`` respectively,
        where each is a pair ``(lower, upper)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param octave_range: the (inclusive) range of octaves (pair of integers)
        :param dtype: dtype of the resulting array
        :return: a one-hot matrix (numpy array)
        """
        flow, fhigh = fifth_range
        olow, ohigh = octave_range
        f = self.fifths()
        o = self.octaves()
        if f < flow or f > fhigh:
            raise ValueError(f"The pitch {self} is outside the given fifth range {fifth_range}.")
        if o < olow or o > ohigh:
            raise ValueError(f"The pitch {self} is outside the given octave range {octave_range}.")
        out = np.zeros((fhigh-flow+1, ohigh-olow+1), dtype=dtype)
        out[f-flow, o-olow] = 1
        return out


@Spelled.link_interval_type()
class SpelledInterval(Spelled, AbstractSpelledInterval, Interval, Diatonic, Chromatic):
    """
    Represents a spelled interval.
    """
    def __init__(self, value):
        """
        Takes a string consisting of the form
        ``-?<quality><generic-size>:<octaves>``,
        e.g. ``"M6:0"``, ``"-m3:0"``, or ``"aa2:1"``,
        which stand for a major sixth, a minor third down, and a double-augmented ninth, respectively.
        possible qualities are d (diminished), m (minor), M (major), P (perfect), and a (augmented),
        where d and a can be repeated.

        :param value: a string or internal numeric representation of the interval
        """
        if isinstance(value, str):
            sign, octaves, fifths = self.parse_interval(value)
            assert isinstance(sign, numbers.Integral)
            assert isinstance(octaves, numbers.Integral)
            assert isinstance(fifths, numbers.Integral)
            assert abs(sign) == 1
            assert octaves >= 0
            # correct octaves from fifth steps
            octaves -= Spelled.diatonic_steps_from_fifths(fifths) // 7
            value = np.array([octaves, fifths])
            # negate value for negative intervals
            if sign < 0:
                value *= -1
        else:
            octaves, fifths = value
            assert isinstance(octaves, numbers.Integral)
            assert isinstance(fifths, numbers.Integral)
            value = np.array([octaves, fifths])
        super().__init__(value=value, is_pitch=False, is_class=False)

    @staticmethod
    def from_fifths_and_octaves(fifths, octaves):
        """
        Create an interval by directly providing its internal fifths and octaves.

        :param fifths: the fifths (= interval class) of the interval (integer)
        :param octaves: the internal/dependent octaves of the interval (integer)
        :return: the resulting interval (SpelledInterval)
        """
        return SpelledInterval((octaves, fifths))

    @staticmethod
    def from_independent(fifths, octaves):
        """
        Create an interval from fifths (interval class) and independent/external octaves.

        :param fifths: the fifth (= interval class) of the interval (integer)
        :param octaves: the external/independent octaves the interval spans (integer)
        :return: the resulting interval (SpelledInterval)
        """
        return SpelledInterval.from_fifths_and_octaves(fifths, octaves - (fifths * 4) // 7)

    @staticmethod
    def from_onehot(onehot, fifth_low, octave_low):
        """
        Create a spelled interval from a one-hot matrix.

        Requires the lower bounds of the fifth and octave range used by the one-hot matrix.

        :param onehot: a one-hot matrix representing the interval (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot matrix
        :param octave_low: the lowest octave expressible in the one-hot matrix
        :return: the resulting interval (SpelledInterval)
        """
        if onehot.sum() != 1:
            raise ValueError(f"{onehot} is not a one-hot vector.")
        fs, os = np.where(onehot==1)
        return SpelledInterval.from_independent(fs[0] + fifth_low, os[0] + octave_low)

    # interval interface

    @classmethod
    def unison(cls):
        """
        Create a perfect unison.

        :return: P1:0
        """
        return cls.from_fifths_and_octaves(0,0)

    @classmethod
    def octave(cls):
        """
        Create a perfect octave.

        :return: P1:1
        """
        return cls.from_fifths_and_octaves(0,1)
    
    @classmethod
    def chromatic_semitone(cls):
        """
        Create a chromatic semitone.

        :return: a1:0
        """
        return SpelledInterval.from_fifths_and_octaves(7,-4)

    def direction(self):
        """
        Returns the direction of the interval (1 = up, 0 = neutral, -1 = down).
        Only perfect unisons are considered neutral.

        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        ds = self.diatonic_steps()
        if ds == 0:
            alt = (self.fifths() + 1) // 7
            if alt == 0:
                return 0
            elif alt < 0:
                return -1
            else:
                return 1
        elif ds < 0:
            return -1
        else:
            return 1

    def __abs__(self):
        if self.direction() < 0:
            return -self
        else:
            return self

    def to_class(self):
        return self.IntervalClass(self.value[1])

    def ic(self):
        return self.to_class()

    def embed(self):
        return self

    def is_step(self):
        return abs(self.diatonic_steps()) <= 1

    # spelled interface

    def name(self):
        octave = abs(self.octaves())
        if self.direction() == -1:
            # negative intervals are to be printed with "-" sign
            sign = "-"
            # in return we have to invert the interval class
            inverse = True
            # in the interval representation, the octave "0" is positive, while "-1" is the first negative octave;
            # an octave of "-1" in internal representation (i.e. the first negative octave) therefore corresponds to an
            # octave "0" with negative sign in printing; we thus need to subtract one from the absolute value for
            # printing; the unison (perfect, diminished or augmented) are an exception because they correspond to
            # zero diatonic steps (when ignoring the octave)
            if self.diatonic_steps() % 7 != 0:
                octave -= 1
        else:
            sign = ""
            inverse = False
        return sign + self.interval_class_from_fifths(self.fifths(), inverse=inverse) + f":{octave}"

    def compare(self, other):
        """
        Comparison between two spelled intervals according to diatonic ordering.

        Returns 0 if the intervals are equal,
        1 if the first interval (``self``) is greater,
        and -1 if the second interval (``other``) is greater.

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).

        :param other: an interval to compare to (SpelledInterval)
        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledInterval):
            return (self-other).direction()
        else:
            raise TypeError(f"Cannot compare {type(self)} with {type(other)}.")

    def fifths(self):
        return self.value[1]

    def octaves(self):
        return self.value[0] + (self.fifths() * 4) // 7

    def internal_octaves(self):
        return self.value[0]

    def generic(self):
        if self.direction() < 0:
            return -(-self).degree()
        else:
            return self.degree()

    def diatonic_steps(self):
        return (self.fifths() * 4) + (self.internal_octaves() * 7)

    def alteration(self):
        return (abs(self).fifths() + 1) // 7

    def onehot(self, fifth_range, octave_range, dtype=int):
        """
        Returns a one-hot encoding of the interval in fifths (first dimension) and independent octaves (second dimension).
        The range of fifths and octaves is given by ``fifth_range`` and ``octave_range`` respectively,
        where each is a tuple ``(lower, upper)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param octave_range: the (inclusive) range of octaves (pair of integers)
        :param dtype: dtype of the resulting array
        :return: a one-hot matrix (numpy array)
        """
        flow, fhigh = fifth_range
        olow, ohigh = octave_range
        f = self.fifths()
        o = self.octaves()
        if f < flow or f > fhigh:
            raise ValueError(f"The interval {self} is outside the given fifth range {fifth_range}.")
        if o < olow or o > ohigh:
            raise ValueError(f"The interval {self} is outside the given octave range {octave_range}.")
        out = np.zeros((fhigh-flow+1, ohigh-olow+1), dtype=dtype)
        out[f-flow, o-olow] = 1
        return out


@Spelled.link_pitch_class_type()
class SpelledPitchClass(Spelled, AbstractSpelledPitch, Pitch):
    """
    Represents a spelled pitch class, i.e. a pitch without octave information.
    """
    def __init__(self, value):
        """
        Takes a string consisting of the form
        ``<letter><accidentals?>``, e.g. ``"C#"``, ``"E"``, or ``"Dbb"``.
        Accidentals may be written as ASCII symbols (#/b)
        or with unicode symbols (♯/♭), but not mixed within the same note.

        :param value: a string or internal numeric representation of the pitch class
        """
        if isinstance(value, str):
            octaves, fifths = self.parse_pitch(value)
            assert octaves is None
        else:
            fifths = value
        assert isinstance(fifths, numbers.Integral)
        super().__init__(value=fifths, is_pitch=True, is_class=True)

    @staticmethod
    def from_fifths(fifths):
        """
        Create a pitch class by directly providing its position on the line of fifths (C=0, G=1, D=2, ...).

        :param fifths: the line-of-fifths position of the pitch class (integer)
        :return: the resulting pitch class (SpelledPitchClass)
        """
        return SpelledPitchClass(fifths)

    @staticmethod
    def from_onehot(onehot, fifth_low):
        """
        Create a spelled pitch class from a one-hot vector.

        Requires the lower bounds of the fifth range used by the one-hot vector.

        :param onehot: a one-hot vector representing the pitch (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot vector
        :return: the resulting pitch class (SpelledPitchClass)
        """
        if onehot.sum() != 1:
            raise ValueError(f"{onehot} is not a one-hot vector.")
        fs, = np.where(onehot==1)
        return SpelledPitchClass.from_fifths(fs[0] + fifth_low)

    # pitch interface

    def interval_from(self, other):
        if type(other) == SpelledPitchClass:
            return SpelledIntervalClass.from_fifths(self.value-other.value)
        else:
            raise TypeError(f"Cannot take interval between SpelledPitchClass and {type(other)}.")

    def pc(self):
        return self

    def embed(self):
        return SpelledPitch.from_fifths_and_octaves(self.fifths(), -((self.fifths() * 4) // 7))

    # spelled interface

    def name(self):
        return self.pitch_class_from_fifths(self.fifths())
    
    def compare(self, other):
        """
        Comparison between two spelled pitch classes according to line-of-fifth.

        Returns 0 if the pitch classes are equal,
        1 if the first pitch class (``self``) is greater ("sharper"),
        and -1 if the second pitch class (``other``) is greater ("sharper").

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).

        :param other: a pitch class to compare to (SpelledPitchClass)
        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledPitchClass):
            return np.sign(self.fifths() - other.fifths())
        else:
            raise TypeError(f"Cannot compare {type(self)} with {type(other)}.")

    def fifths(self):
        return self.value

    def octaves(self):
        return 0

    def internal_octaves(self):
        return 0

    def alteration(self):
        return (self.fifths() + 1) // 7

    def letter(self):
        return chr(ord('A') + (self.degree() + 2) % 7)

    def onehot(self, fifth_range, dtype=int):
        """
        Returns a one-hot encoding of the pitch class in fifths.
        The range of fifths is given by ``fifth_range`` as a tuple ``(lower, upper)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param dtype: dtype of the resulting array
        :return: a one-hot vector (numpy array)
        """
        low, high = fifth_range
        f = self.fifths()
        if f < low or f > high:
            raise ValueError(f"The pitch class {self} is outside the given fifths range {fifth_range}.")
        out = np.zeros(high-low+1, dtype=dtype)
        out[f-low] = 1
        return out


@Spelled.link_interval_class_type()
class SpelledIntervalClass(Spelled, AbstractSpelledInterval, Interval, Diatonic, Chromatic):
    """
    Represents a spelled interval class, i.e. an interval without octave information.
    """
    def __init__(self, value):
        """
        Takes a string consisting of the form
        ``-?<quality><generic-size>``,
        e.g. ``"M6"``, ``"-m3"``, or ``"aa2"``,
        which stand for a major sixth, a minor third down (= major sixth up), and a double-augmented second, respectively.
        possible qualities are d (diminished), m (minor), M (major), P (perfect), and a (augmented),
        where d and a can be repeated.

        :param value: a string or internal numeric representation of the interval class
        """
        if isinstance(value, str):
            sign, octaves, fifths = self.parse_interval(value)
            assert isinstance(sign, numbers.Integral)
            assert abs(sign) == 1
            assert octaves is None
            assert isinstance(fifths, numbers.Integral)
            fifths *= sign
        else:
            fifths = value
        assert isinstance(fifths, numbers.Integral)
        super().__init__(value=fifths, is_pitch=False, is_class=True)

    @staticmethod
    def from_fifths(fifths):
        """
        Create an interval class by directly providing its internal fifths.

        :param fifths: the line-of-fifths position of the interval class (integer)
        :return: the resulting interval class (SpelledIntervalClass)
        """
        return SpelledIntervalClass(fifths)

    @staticmethod
    def from_onehot(onehot, low):
        """
        Create a spelled interval class from a one-hot vector.

        Requires the lower bounds of the fifth range used by the one-hot vector.

        :param onehot: a one-hot vector representing the interval (numpy array)
        :param fifth_low: the lowest fifth expressible in the one-hot vector
        :return: the resulting interval class (SpelledIntervalClass)
        """
        if onehot.sum() != 1:
            raise ValueError(f"{onehot} is not a one-hot vector.")
        fs, = np.where(onehot==1)
        return SpelledIntervalClass.from_fifths(fs[0] + low)

    # interval interface

    @classmethod
    def unison(cls):
        """
        Return a perfect unison.

        :return: P1
        """
        return cls.from_fifths(0)

    @classmethod
    def octave(cls):
        """
        Return a perfect unison, which is the same as an octave for interval classes.

        :return: P1
        """
        return cls.from_fifths(0)

    @classmethod
    def chromatic_semitone(cls):
        """
        Return a chromatic semitone

        :return: a1
        """
        return SpelledIntervalClass.from_fifths(7)

    def direction(self):
        """
        Returns the direction of smallest realization of the interval class
        (1 = up, 0 = neutral, -1 = down).
        For example, the direction of ``M7`` (= ``m2``) is down.

        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        ds = self.degree()
        if ds == 0:
            alt = (self.fifths() + 1) // 7
            if alt == 0:
                return 0
            elif alt < 0:
                return -1
            else:
                return 1
        elif ds > 3:
            return -1
        else:
            return 1

    def __abs__(self):
        if self.direction() < 0:
            return -self
        else:
            return self

    def ic(self):
        return self

    def embed(self):
        return SpelledInterval.from_fifths_and_octaves(self.fifths(), -((self.fifths() * 4) // 7))

    def is_step(self):
        return self.degree() in [0,1,6]

    # spelled interface

    def name(self, inverse=False):
        if inverse:
            sign = "-"
        else:
            sign = ""
        return sign + self.interval_class_from_fifths(self.fifths(), inverse=inverse)
    
    def compare(self, other):
        """
        Comparison between two spelled interval classes according to line-of-fifth.

        Returns 0 if the interval classes are equal,
        1 if the first interval class (``self``) is greater ("sharper"),
        and -1 if the second interval class (``other``) is greater ("sharper").

        This method can be indirectly used through binary comparison operators
        (including ``==``, ``<`` etc.).

        :param other: an interval class to compare to (SpelledIntervalClass)
        :return: ``-1`` / ``0`` / ``1`` (integer)
        """
        if isinstance(other, SpelledIntervalClass):
            return np.sign(self.fifths() - other.fifths())
        else:
            raise TypeError(f"Cannot compare {type(self)} with {type(other)}.")

    def fifths(self):
        return self.value

    def octaves(self):
        return 0

    def internal_octaves(self):
        return 0

    def generic(self):
        return self.degree()

    def diatonic_steps(self):
        return self.degree()

    def alteration(self):
        return (self.fifths() + 1) // 7

    def onehot(self, fifth_range, dtype=int):
        """
        Returns a one-hot encoding of the interval class in fifths.
        The range of fifths is given by ``fifth_range`` as a tuple ``(lower, upper)``.

        :param fifth_range: the (inclusive) range of fifths (pair of integers)
        :param dtype: dtype of the resulting array
        :return: a one-hot vector (numpy array)
        """
        low, high = fifth_range
        f = self.fifths()
        if f < low or f > high:
            raise ValueError(f"The pitch class {self} is outside the given fifths range {fifth_range}.")
        out = np.zeros(high-low+1, dtype=dtype)
        out[f-low] = 1
        return out
